import numpy as np

def gaussian(X, mu, cov):
    """
        Returns the pdf of a Gaussian kernel

        :param X: Input data.
        :type X: :obj:`numpy.ndarray`
        :param mu: Mean of the gaussian distribution.
        :type mu: :obj:`numpy.ndarray`
        :param cov: Covariance matrix.
        :type cov: :obj:`numpy.ndarray`

        :formula:

            .. math:: \\dfrac{1}{(2 \\pi)^{\\frac{n}{2}} \\sqrt{\\det{\\sigma}}} e^{-\\dfrac{1}{2} \\dfrac{X'X}{\\sigma^{2}}}

        :references: * Murphy, K. P. (2012). Machine learning: a probabilistic perspective. MIT press.
    """
    X_centered = X - mu
    n = X.shape[1]
    _a = (1 / ((2 * np.pi) ** (n / 2) * np.linalg.det(cov) ** 0.5))
    _b = np.exp(-0.5 * np.dot(np.dot(X_centered, np.linalg.inv(cov)), X_centered.T))

    return np.diagonal(_a * _b).reshape(-1, 1)



import numpy as np
import matplotlib.pyplot as plt

# TODO: Add priors for the classes for GGM @ybendou

class GGM:

    def __init__(self):
        """
            Gaussian Generative model.
            This class implements a Linear and Quadratic classifiers obtained by assuming Gaussian distributed data and by using Bayes' Theorem.

            .. warning:: This function is still under development.
                This is a beta version, please be aware that some functionalitie might not be available.
                The full stable version soon will be released.
                To be added: priors of the classes

            :formula: .. math:: \\mathbb{P}(A \\mid B) = \\dfrac{\\mathbb{P}(B \\mid A) \\cdot \\mathbb{P}(A)}{\\mathbb{P}(B)}

            :references: * Murphy, K. P. (2012). Machine learning: a probabilistic perspective. MIT press.
        """
        self.class_means = None
        self.cov = None
        self.isotropic = True

    def _weighted_norm(self, a, cov=None, norm="euclidian"):
        """
            Returns the norm of a vector using the euclidian or the mahalonobis norm.

            :param cov: Covariance matrix.
            :type  cov: :obj:`numpy.ndarray`
            :param norm: Norm to be used, options are euclidian or mahalonobis, if mahalonobis is specified a proper covariance matrix
                has to be specified too, default is euclidian.
            :type  norm: :obj:`str`
        """
        if norm == "euclidian":
            return np.sum(a**2)
        elif norm == "mahalanobis":
            # assert cov != None, "Specify a covariance matrix"
            return np.dot(np.dot(a, np.linalg.inv(cov)), a)

    def _compute_distance(self, new_data, norm="euclidian"):
        """
            Returns distance between each point and the mean of each class, picks the one that maximizes the likelihood.

            :param new_data: New data features.
            :type  new_data: :obj:`numpy.ndarray`
            :param norm: Norm to be used, options are 'euclidian' or 'mahalonobis', default is 'euclidian'.
            :type  norm: :obj:`str`
        """
        nb_classes = self.class_means.shape[0]
        distances = np.zeros((new_data.shape[0], nb_classes))

        for c in range(nb_classes):  # per class
            if self.isotropic:
                # Add the prior later, ln(p(Ck))
                distances[:, c] = np.apply_along_axis(self._weighted_norm, 1, new_data - self.class_means[c, :], cov=self.cov, norm=norm)
            else:
                distances[:, c] = -np.apply_along_axis(self._weighted_norm, 1, new_data - self.class_means[c, :], cov=self.cov[c, :, :], norm=norm)
                distances[:, c] -= (1 / 2) * np.log(np.linalg.det(self.cov[c, :, :]))
        return distances

    def fit(self, data, labels, nb_classes, isotropic=True):
        """
            Extracts the mean and the covariance of each class, in the isotropic case we assume that all the classes have the same covariance.
            Returns the mean vector for each class, and the estimated covariance.

            :param data: Data features.
            :type  data: :obj:`numpy.ndarray`
            :param labels: Data labels.
            :type  labels: :obj:`numpy.ndarray`
            :param nb_classes: Number of classes in the data.
            :type  nb_classes: :obj:`int`
            :param isotropic: Is an isotropic case or not, meaning the covariance matrix for each class
                is a :math:`(\\sigma^{2}) \\times \\mathbb{I}_{n}` or different (LDA vs QDA).
            :type isotropic: :obj:`bool`
        """
        self.isotropic = isotropic
        class_means = np.zeros((nb_classes, data.shape[1]))

        if not self.isotropic:
            cov = np.zeros((nb_classes, data.shape[1], data.shape[1]))
        else:
            cov = np.cov(data.T)

        for c in range(nb_classes):
            class_means[c, :] = data[labels == c].mean(axis=0)
            if not self.isotropic:
                cov[c, :, :] = np.cov(data[labels == c].T)

        self.class_means = class_means
        self.cov = cov
        return class_means, cov

    def predict(self, new_data, norm="euclidian"):
        """
            Returns predictions for each sample, it affects the labels by finding the closest mean of each class
            using different distances (euclidian, mahalanobis with isotropic or non isotropic covariance)
            For the isotropic case we use a Linear Discriminant classifier, for the non isotropic case we use a
            Quadratic Discriminant classifier.

            :param new_data: New data to evaluate.
            :type  new_data: :obj:`numpy.ndarray`
            :param norm: Norm to be used, options are euclidian or mahalonobis, default is euclidian.
            :type  norm: :obj:`str`
        """
        distances = self._compute_distance(new_data, norm=norm)

        objective = {True: np.argmin, False: np.argmax}  # Maximize or minimize based on the case of isotropic or non isotropic
        predicted_labels = objective[self.isotropic](distances, axis=1)

        return predicted_labels  # this function classifies each point in the test set by affecting it to the

    def predict_proba(self, X, norm="euclidian"):
        """
            Returns the likelihood probability for each class.

            :param X: Data features.
            :type  X: :obj:`numpy.ndarray`
            :param norm: Norm to be used, options are euclidian or mahalonobis, default is euclidian.
            :type  norm: :obj:`str`
        """
        distances = self._compute_distance(X, norm=norm)
        likelihood = np.exp(distances - distances.max(axis=1)[:, np.newaxis])

        return likelihood / likelihood.sum(axis=1)[:, np.newaxis]

    def plot_decision_boundary(self, X, labels, norm="euclidian", grid_size=100, *args):
        """
            Plots the predictions on the entire dataset as well as the decision boudaries for each class

            :param X: Data features.
            :type  X: :obj:`numpy.ndarray`
            :param labels: Labels of each point in the training set.
            :type  labels: :obj:`numpy.ndarray`
            :param norm: Norm to be used, options are euclidian or mahalonobis, default is euclidian.
            :type  norm: :obj:`str`
            :param grid_size: Size of the square grid to be plotted.
            :type  grid_size: :obj:`int`
        """
        x = np.linspace(np.min(X[:, 0]), np.max(X[:, 0]), grid_size)  # extent of the grid on the x axis
        y = np.linspace(np.min(X[:, 1]), np.max(X[:, 1]), grid_size)  # extent of the grid on the y axis
        [xx, yy] = np.meshgrid(x, y)
        Z = self.predict(np.c_[xx.ravel(), yy.ravel()], norm=norm)
        Z = Z.reshape(grid_size, grid_size)
        plt.figure(*args)
        plt.scatter(X[:, 0], X[:, 1], c=labels)
        plt.contourf(xx, yy, Z, cmap=plt.cm.RdBu, alpha=0.6)
        plt.show()



import numpy as np
import math
from scipy import stats as scp

from . import descriptive as desc
from ..misc import test_summary


# One sample Student test
def ttest(x, mu=0, alpha=0.05, is_bernoulli=False, two_sided=True, return_tuple=False):
    """One sample Student's test.

        In the two-sided setup, we aim at testing:

        .. math:: H_{0}: \\bar{X} = \\mu \\text{  against  } H_{1}: \\bar{X} \\neq \\mu

        The one-sided setup tests :math:`H_{0}: \\bar{X} \\leq \\mu` against :math:`H_{0}: \\bar{X} > \\mu`.

        The p-value is computed as:

        .. math:: \\mathbb{P}(|Z| \\geq |t| \\mid H_{0} \\text{ holds})

        with, under :math:`H_{0}`:

        .. math:: t = \\dfrac{\\bar{X}_{1} - \\mu }{ \\dfrac{s}{\\sqrt{n}} } \\sim \\mathcal{N}(0, 1)

        if :math:`s = \\mathbb{V}(\\mathbf{X})` is known or if :math:`n \\gg 30`, otherwise :math:`t \\sim \\mathcal{T}_{n - 1}`,
        a Student distribution with :math:`n-1` degrees of freedom.
        In case of a bernoulli distribution, then :math:`s = \\mu(1 - \\mu)`.

    :param x: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x: :obj:`numpy.array`
    :param mu: Theoretical mean to be evaluated in the null hypothesis, defaults to 0.
    :type mu: :obj:`int`, optional
    :param alpha: Confidence level, defaults to 0.05.
    :type alpha: :obj:`float`, optional
    :param is_bernoulli: Input value follows a Bernoulli distribution, i.e. :math:`\\mathbf{X} \\sim \\mathcal{B}(p)` with :math:`p \\in [0, 1]`, defaults to False.
    :type is_bernoulli: :obj:`bool`, optional
    :param two_sided: Perform a two-sided test, defaults to True.
    :type two_sided: :obj:`bool`, optional
    :param return_tuple: Return a tuple with t statistic, critical value and p-value, defaults to False.
    :type return_tuple: :obj:`bool`, optional

    :example:

    >>> from statinf import stats
    >>> stats.ttest([30.02, 29.99, 30.11, 29.97, 30.01, 29.99], mu=30)
    ... +------------------------------------------------------------+
    ... |                   One Sample Student test                  |
    ... +------------+----------------+------------+---------+-------+
    ... |     df     | Critical value |    T-value | p-value |   H0  |
    ... +------------+----------------+------------+---------+-------+
    ... |          5 |   2.5705818366 |  0.7392961 | 0.49295 | True  |
    ... +------------+----------------+------------+---------+-------+
    ... * Confidence level is 95.0%, we need p > 0.025 for two-sided test
    ... * We cannot reject the hypothesis H0: X_bar = 30

    :reference: * DeGroot, M. H., & Schervish, M. J. (2012). Probability and statistics. Pearson Education.
        * Student. (1908). The probable error of a mean. Biometrika, 1-25.

    :return: Summary for the test or tuple statistic, critical value, p-value.
    :rtype: :obj:`str` or :obj:`tuple`
    """

    # Define test degrees of freedom
    if two_sided:
        quant_order = 1 - (alpha / 2)
        h0 = f'X_bar = {mu}'
        h1 = f'X_bar != {mu}'
    else:
        quant_order = 1 - alpha
        h0 = f'X_bar <= {mu}'
        h1 = f'X_bar > {mu}'

    # Input vector as array
    x = np.asarray(x)
    # Sample size
    n = len(x)

    # Empirical mean
    x_bar = x.mean()
    # s estimator (variance)
    if is_bernoulli:
        s2 = x_bar * (1 - x_bar)
    else:
        s2 = desc.var(x)

    # Degrees of freedom
    df = n - 1

    # T statistic
    t = (x_bar - mu) / (math.sqrt(s2 / n))
    if two_sided:
        t = math.fabs(t)
    # p and critical values
    p = 2.0 * (1.0 - scp.t.cdf(t, df=df))

    if n > 30:
        cv = scp.norm.ppf(quant_order)
    else:
        cv = scp.t.ppf(quant_order, df=df)

    _summ = test_summary(df=df, critical_value=cv, t_value=t,
                         p_value=p,
                         title='One Sample Student test',
                         h0=h0, h1=h1,
                         alpha=alpha)

    if return_tuple:
        return t, cv, p
    else:
        return _summ


# 2 samples Student test
def ttest_2samp(x1, x2, alpha=0.05, paired=False, is_bernoulli=False, two_sided=True, return_tuple=False):
    """Two samples Student's test.

        If the samples are independent (:math:`X_{1} \\bot X_{2}`), we test:

        .. math:: H_{0}: \\bar{X}_{1} = \\bar{X}_{2} \\text{  against  } H_{1}: \\bar{X}_{1} \\neq \\bar{X}_{2}

        The test statistic is:

        .. math:: t = \\dfrac{\\bar{X}_{1} - \\bar{X}_{2} }{ \\sqrt{ \\dfrac{(n_{1} - 1) s^{2}_{1} + (n_{2} - 1) s^{2}_{2}}{n_{1} + n_{2} - 1} } \\sqrt{\\dfrac{1}{n_{1}} + \\dfrac{1}{n_{2}}} }

        where:

            * :math:`t \\sim \\mathcal{T}_{n_{1} + n_{2} - 2}`, if :math:`\\mathbf{X} \\sim \\mathcal{N}(\\mu_{X}, \\sigma^{2}_{X})` and :math:`\\mathbf{Y} \\sim \\mathcal{N}(\\mu_{Y}, \\sigma^{2}_{Y})`
            * :math:`t \\sim \\mathcal{N}(0, 1)`, if :math:`n_{1} \\gg 30` and :math:`n_{2} \\gg 30`

        If samples are paired:

        .. math:: H_{0}: \\bar{X}_{1} = \\bar{X}_{2} \\Leftrightarrow \\bar{X}_{1} - \\bar{X}_{2} = 0 \\Leftrightarrow \\bar{X}_{D} = 0

        We then compared a one sample test where the tested vector is the difference between both vectors :math:`X_{D} = X_{1} - X_{2}`
        for which we compare whether it is equal to 0.

    :param x1: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x1: :obj:`numpy.array`
    :param x2: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x2: :obj:`numpy.array`
    :param alpha: Confidence level, defaults to 0.05.
    :type alpha: :obj:`float`, optional
    :param paired: Performs one sample test of the difference if samples are paired, defaults to False.
    :type paired: :obj:`bool`, optional
    :param is_bernoulli: Input variables follow a Bernoulli distribution, i.e. :math:`\\mathbf{X} \\sim \\mathcal{B}(p)` with :math:`p \\in [0, 1]`, defaults to False.
    :type is_bernoulli: :obj:`bool`, optional
    :param two_sided: Perform a two-sided test, defaults to True.
    :type two_sided: :obj:`bool`, optional
    :param return_tuple: Return a tuple with t statistic, critical value and p-value, defaults to False.
    :type return_tuple: :obj:`bool`, optional

    :example:

    >>> from statinf import stats
    >>> a = [30.02, 29.99, 30.11, 29.97, 30.01, 29.99]
    >>> b = [29.89, 29.93, 29.72, 29.98, 30.02, 29.98]
    >>> stats.ttest_2samp(a, b)
    ... +------------------------------------------------------------+
    ... |                   Two Samples Student test                 |
    ... +------------+----------------+------------+---------+-------+
    ... |     df     | Critical value |    T-value | p-value |   H0  |
    ... +------------+----------------+------------+---------+-------+
    ... |         10 |  -1.8124611228 |  1.1310325 | 0.28444 | True  |
    ... +------------+----------------+------------+---------+-------+
    ...  * Confidence level is 95.0%, we need p > 0.05
    ...  * We cannot reject the hypothesis H0: X1 = X2
    ...  * Samples with unequal variances
    ...  * Same sample sizes

    :reference: * DeGroot, M. H., & Schervish, M. J. (2012). Probability and statistics. Pearson Education.
        * Student. (1908). The probable error of a mean. Biometrika, 1-25.

    :return: Summary for the test or tuple statistic, critical value, p-value.
    :rtype: :obj:`str` or :obj:`tuple`
    """
    x = np.asarray(x1)
    y = np.asarray(x2)

    # Define test degrees of freedom
    if two_sided:
        quant_order = 1 - (alpha / 2)
        h0 = 'X1_bar = X2_bar'
        h1 = 'X1_bar != X2_bar'
    else:
        quant_order = 1 - alpha
        h0 = 'X1 <= X2'
        h1 = 'X1 > X2'

    # Sample sizes
    n1, n2 = len(x), len(y)

    if paired:
        # If samples are paired, we perform a 1-sample student test
        # We compare if the difference is different from 0.
        mean1, mean2 = x.mean(), y.mean()
        d = x - y
        t, cv, p = ttest(d, alpha=alpha, return_tuple=True)
        df = len(d)
    else:
        # Else samples are independent
        # Compute means
        mean1, mean2 = x.mean(), y.mean()
        # Compute standard deviations
        if is_bernoulli:
            s1 = mean1 * (1 - mean1)
            s2 = mean2 * (1 - mean2)
        else:
            s1 = desc.var(x)
            s2 = desc.var(y)
        # Compute grouped variance
        sd = np.sqrt(((n1 - 1) * s1 + (n2 - 1) * s2) / (n1 + n2 - 2))
        # Degrees of freedom
        df = n1 + n2 - 2
        # Calculate the t statistic
        t = (mean1 - mean2) / sd

        # calculate the critical value
        cv = scp.t.ppf(quant_order, df)
        # calculate the p-value
        if (n1 > 30) & (n2 > 30):
            p = 2.0 * (1.0 - scp.norm.cdf(math.fabs(t)))
        else:
            p = 2.0 * (1.0 - scp.t.cdf(math.fabs(t), df=df))

    extra = f" * E(X1) = {round(mean1, 3)} and E(X2) = {round(mean2, 3)} \n"
    extra += " * Performed test for paired samples \n" if paired else ''
    extra += " * Large sample sizes, t ~ N(0, 1) from CLT" if (n1 > 30) & (n2 > 30) else ' * Small sample sizes, assumed t ~ T(n-1)'

    _summ = test_summary(df=df, critical_value=cv, t_value=t,
                         p_value=p,
                         title='Two Samples Student test',
                         h0=h0, h1=h1,
                         alpha=alpha,
                         extra=extra)

    if return_tuple:
        return t, cv, p
    else:
        return _summ


# Kolmogorov-Smirnov test
def kstest(x1, x2='normal', alpha=0.05, return_tuple=False, **kwargs):
    """Kolmogorov-Smirnov test for sample tests.

        We test if two samples come from the same distribution :math:`H_{0}: F(x) = G(x)` against :math:`H_{1}: F(x)` follows a distribution that :math:`G(x)`.

        The statistic is:

        .. math:: D_{mn} = \\sup_{-\\infty < x < + \\infty} |F_{n}(x) - G_{m}(x)|

        and the critical value is given by:

        .. math:: c = \\mathbb{P} \\left[ \\left( \\dfrac{mn}{n + m} \\right)^{2} D_{mn} < K_{\\alpha} \\right]

        where :math:`K_{\\alpha}` represents the quantile of order :math:`\\alpha` from a Kolmogorov distribution.

        This test is an alternative to the :math:`\\chi^{2}`-test where we compare two distributions.
        By comparing an unknown empirical distribution :math:`F_{n}` with a distribution :math:`G_{n}` pulled from a known function, we can assess whether :math:`F_{n}` follows the same distribution as :math:`G_{n}`.
        For instance, when comparing :math:`F_{n}` (unknown) to :math:`G_{n} \\sim \\mathcal{N}(\\mu, \\sigma^{2})`,
        not rejecting :math:`H_{0}` would mean that we cannot reject the hypothesis that :math:`F_{n} \\sim \\mathcal{N}(\\mu, \\sigma^{2})`.

    :param x1: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x1: :obj:`numpy.array`
    :param x2: Sample to be compared. Can be external empirical sample in the same format as :obj:`x1` or the name of a cdf which can be 'normal', 'beta', 'gamma', 'poisson', 'chisquare', 'exponential', 'gamma', defaults to 'normal'.
    :type x2: :obj:`str`, optional
    :param alpha: Confidence level, defaults to 0.05.
    :type alpha: :obj:`float`, optional
    :param return_tuple: Return a tuple with K statistic, critical value and p-value, defaults to False.
    :type return_tuple: :obj:`bool`, optional

    :example:

    >>> from statinf import stats
    >>> import numpy as np
    >>> stats.kstest(np.random.normal(size=100))
    ... +------------------------------------------------------------+
    ... |                   Kolmogorov-Smirnov test                  |
    ... +------------+----------------+------------+---------+-------+
    ... | D value    | Critical value | K-value    | p-value |   H0  |
    ... +------------+----------------+------------+---------+-------+
    ... |       0.09 |  1.35809863932 |  0.6363961 | 0.81275 | True  |
    ... +------------+----------------+------------+---------+-------+
    ...  * We cannot reject the hypothesis H0: F(x) ~ normal
    ...  * Confidence level is 95.0%, we need p > 0.05

    :reference: * DeGroot, M. H., & Schervish, M. J. (2012). Probability and statistics. Pearson Education.
        * Kolmogorov-Smirnov, A. N., Kolmogorov, A., & Kolmogorov, M. (1933). Sulla determinazione empírica di uma legge di distribuzione.
        * Marsaglia, G., Tsang, W. W., & Wang, J. (2003). `Evaluating Kolmogorov’s distribution <https://www.jstatsoft.org/article/view/v008i18>`_. Journal of Statistical Software, 8(18), 1-4.

    :return: Summary for the test or tuple statistic, critical value, p-value.
    :rtype: :obj:`str` or :obj:`tuple`
    """

    quant_order = 1 - alpha
    # Format x1
    x = np.asarray(x1)
    n = len(x)

    if type(x2) == str:
        h0 = f"F(x) ~ {x2}"
        h1 = f"F(x) is not {x2}"
        if x2.lower() in ['normal', 'norm', 'gaussian', 'gauss']:
            y = np.random.normal(size=n, **kwargs)
        elif x2.lower() in ['beta']:
            y = np.random.beta(size=n, **kwargs)
        elif x2.lower() in ['gamma']:
            y = np.random.gamma(size=n, **kwargs)
        elif x2.lower() in ['poisson']:
            y = np.random.poisson(size=n, **kwargs)
        elif x2.lower() in ['chisquare', 'chi-square', 'chi2', 'x2']:
            y = np.random.chisquare(size=n, **kwargs)
        elif x2.lower() in ['exponential', 'expo']:
            y = np.random.exponential(size=n, **kwargs)
        elif x2.lower() in ['gumbel']:
            y = np.random.gumbel(size=n, **kwargs)
        else:
            raise ValueError('cdf value for x2 is not valid.')
    else:
        y = np.asarray(x2)
        h0 = "F(x) = G(x)"
        h1 = "F(x) and G(x) have different distributions"

    # Sample sizes
    m = len(y)
    # Sort values and concatenate to one array
    x_sort = np.sort(x)
    y_sort = np.sort(y)
    concat_data = np.concatenate([x_sort, y_sort])
    # Build sort distributions to compute the max deviance
    cdf1 = np.searchsorted(x_sort, concat_data, side='right') / float(n)
    cdf2 = (np.searchsorted(y_sort, concat_data, side='right')) / float(m)
    # Find the max deviance
    d = np.max(np.absolute(cdf1 - cdf2))
    # Compute K value
    k = d * math.sqrt((n * m) / (n + m))
    cv = scp.kstwobign.ppf(quant_order)
    p = 1.0 - scp.kstwobign.cdf(math.fabs(k))

    _summ = test_summary(df=n, critical_value=cv, t_value=k,
                         p_value=p, alpha=alpha,
                         title='Kolmogorov-Smirnov test',
                         h0=h0, h1=h1,
                         extra=f' * The D-value is: {round(d, 5)}')

    if return_tuple:
        return k, cv, p
    else:
        return _summ



import numpy as np
from ..data import rankdata


# ##### Variance

def var(x, std=False, df=1):
    """Compute the variance of a variable.

    :param x: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x: :obj:`numpy.array`
    :param std: Returns standard deviation, i.e. :math:`\\sqrt{\\mathbb{V}(\\mathbf{X})}`, defauls to False.
    :type std: :obj:`bool`, optional
    :param df: Degrees of freedom, defaults to 1.
    :type df: :obj:`int`, optional

    :formula: .. math:: \\mathbb{V}(\\mathbf{X}) = \\dfrac{1}{n - 1} \\sum_{i = 1}^{n} (X_{i} - \\bar{X})^{2}

    :example:

    >>> from statinf import stats
    >>> x = [0.023699, 0.021436, 0.0200109, 0.0202762, 0.0165271, 0.01027]
    >>> stats.var(x)
    ... 2.2492979044000003e-05

    :return: Variance.
    :rtype: :obj:`float`
    """
    x = np.asarray(x)
    n = len(x)
    x_bar = x.mean()
    s2 = ((x - x_bar) ** 2).sum() / (n-df)
    if std:
        return np.sqrt(s2)
    else:
        return s2


# ##### Covariance

def cov(x, y):
    """Compute the covariance between two variables.

    :param x: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x: :obj:`numpy.array`
    :param y: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type y: :obj:`numpy.array`

    :formula: .. math:: Cov(\\mathbf{X}, \\mathbf{Y}) = \\dfrac{ \\sum_{i = 1}^{n} (X_{i} - \\bar{X}) (Y_{i} - \\bar{Y}) }{n - 1}

    :example:

    >>> from statinf import stats
    >>> x = [0.023699, 0.021436, 0.0200109, 0.0202762, 0.0165271, 0.01027]
    >>> y = [9.4228, 9.27951, 9.167963, 9.68820, 9.56490, 7.543]
    >>> stats.cov(x, y)
    ... 0.003047229298620001

    :reference: * DeGroot, M. H., & Schervish, M. J. (2012). Probability and statistics. Pearson Education.

    :return: Covariance value.
    :rtype: :obj:`float`
    """
    x = np.asarray(x)
    y = np.asarray(y)
    x - x.mean()
    y - y.mean()

    x_xbar = x - x.mean()
    y_ybar = y - y.mean()

    num = (x_xbar * y_ybar).sum()
    cov = num / (len(x) - 1)
    return cov


# ##### Pearson's correlation coefficient

def pearson(x, y):
    """Compute the Pearson's coefficient correlation

    :param x: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x: :obj:`numpy.array`
    :param y: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type y: :obj:`numpy.array`

    :formula: .. math:: \\rho = \\dfrac{ Cov(X, Y) }{\\sigma_{X} \\sigma_{Y}}

        where :math:`\\sigma_{Z} = \\sqrt{\\mathbb{V}(Z)}`

    :example:

    >>> from statinf import stats
    >>> x = [0.023699, 0.021436, 0.0200109, 0.0202762, 0.0165271, 0.01027]
    >>> y = [9.4228, 9.27951, 9.167963, 9.68820, 9.56490, 7.543]
    >>> stats.pearson(x, y)
    ... 0.9750052703452801

    :reference: * DeGroot, M. H., & Schervish, M. J. (2012). Probability and statistics. Pearson Education.

    :return: Pearson's coefficient correlation.
    :rtype: :obj:`float`
    """
    x = np.asarray(x)
    y = np.asarray(y)
    cv = cov(x, y)
    stdx = x.std()
    stdy = y.std()
    return cv / (stdx * stdy)


# ##### Spearman's rank correlation coefficient

def spearman(x, y):
    """Spearman's rank correlation coefficient.

    :param x: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x: :obj:`numpy.array`
    :param y: Input variable. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type y: :obj:`numpy.array`

    :formula: .. math:: \\rho = 1 - \\dfrac{ 6 \\sum_{i=1}^{n} d^{2} }{ n (n^{2} - 1)}

    :example:

    >>> from statinf import stats
    >>> x = [0.023699, 0.021436, 0.0200109, 0.0202762, 0.0165271, 0.01027]
    >>> y = [9.4228, 9.27951, 9.167963, 9.68820, 9.56490, 7.543]
    >>> stats.spearman(x, y)
    ... 0.37142857142857144

    :return: Spearman's rank correlation coefficient.
    :rtype: :obj:`float`
    """
    n = len(x)
    rk_x = rankdata(x)
    rk_y = rankdata(y)
    d = (rk_y - rk_x) ** 2

    rho = 1 - (6 * d.sum()) / (n * (n ** 2 - 1))

    return rho



import numpy as np
import pandas as pd
import math
import warnings
import scipy.stats as scp
import matplotlib.pyplot as plt

from ..ml.initializations import init_params
from ..ml.activations import logit

from ..data.ProcessData import parse_formula
from ..misc import ConvergenceWarning, summary, get_significance
# from .glm_test import GLM

# TODO: HC covariance
# TODO: check probit

class GLM:

    def __init__(self, formula, data, family='binomial', fit_intercept=False, initializer='zeros'):
        """Generalized Linear Model implemented with Newton-Raphson's method.

        :param formula: Regression formula to be run of the form :obj:`y ~ x1 + x2`. See :py:meth:`statinf.data.ProcessData.parse_formula`.
        :type formula: :obj:`str`
        :param data: Input data with Pandas format.
        :type data: :obj:`pandas.DataFrame`
        :param family: Family distribution of the dependent variable, defaults to binomial.
        :type family: :obj:`str`, optional
        :param fit_intercept: Used for adding intercept in the regression formula, defaults to False.
        :type fit_intercept: :obj:`bool`, optional
        :param initializer: Method for initializing the first parameters (see :py:meth:`statinf.ml.initializations`), defaults to 'zeros'.
        :type initializer: :obj:`str`, optional

        .. note:: The modules allows Binomial (for Logit) and Gaussian (for Probit) distributions.
            It will soon be extended to other distributions.
        """
        # Parse the formula
        self.df, self.X_col, self.Y_col = parse_formula(data=data, formula=formula, check_values=True, return_all=True)
        self.formula        = formula
        self.fit_intercept  = fit_intercept
        self.n              = self.df.shape[0]
        self.p              = len(self.X_col) + 1 if fit_intercept else len(self.X_col)
        self.X              = self.df[self.X_col].to_numpy()
        self.Y              = np.array(self.df[self.Y_col].values).reshape(self.n, 1)
        self.cov_type       = 'nonrobust'
        self.family         = family
        self.log_likelihood = np.Inf
        self.log_likelihood_hist = []
        self.gradient_hist  = []
        self.variance_hist  = []
        self.coefs_hist     = []
        self.hessian_hist   = []

        # Initialize parameters
        self.beta = init_params(rows=self.p, cols=1, method=initializer, tensor=False)

    def _get_X(self, new_data=None):
        """Function to process data (and append unit variable if fit_intercept)

        :param new_data: Data to transform, defaults to None (gets train data provided in __init__).
        :type new_data: :obj:`pandas.DataFrame`, optional

        :return: Formatted data
        :rtype: :obj:`numpy.ndarray`
        """
        if new_data is None:
            X = self.X
        else:
            # If new_data, parse the formula to get transformations
            df = parse_formula(data=new_data, formula=self.formula, check_values=True)
            X = df[self.X_col].to_numpy()

        # If fit_intercept, append unit variable
        if self.fit_intercept:
            X = np.hstack((np.ones((self.n, 1), dtype=self.X.dtype), self.X))

        return X

    def _prob(self, X):
        """ Prob
        """
        if self.family == 'binomial':
            prob = logit(X, weights=self.beta)  # (X.dot(self.beta))
        elif self.family == 'gaussian':
            prob = scp.norm.cdf(X.dot(self.beta), loc=0, scale=1)
        else:
            raise ValueError('Family distribution is not valid.')

        return prob

    def _log_likelihood(self, X=None, Y=None):
        """
        Compute Log-Likelihood
        """
        X_data = self._get_X() if X is None else X
        Y_data = self.Y if Y is None else Y

        # Individual log-likelihood
        log_likelihood_id = np.array(Y_data * np.log(self._prob(X_data)) + (1 - Y_data) * np.log(1 - self._prob(X_data)))
        # Overall log-likelihood
        log_likelihood = log_likelihood_id.sum()
        return log_likelihood

    def _gradient(self):
        """
        Gradient (Overall) not the score function
        """
        X = self._get_X()
        gradient = X.T.dot((self._prob(X) - self.Y)) / self.n  # gradient of the likelihood
        return gradient

    def _hessian(self):
        """
        Hessian - Fisher Information
        """
        X = self._get_X()
        pi = self._prob(X)
        S = np.multiply(pi, (1 - pi)) * np.identity(self.n)
        hessian = (1 / self.n) * X.T.dot(S).dot(X)
        return hessian

    def _sandwich(self):
        X = self._get_X()
        pi = self._prob(X)
        res_2 = (self.Y - pi)**2 * np.identity(self.n)
        XX_1 = np.linalg.inv(X.T.dot(X))
        Xres_2X = X.T.dot(res_2.dot(X))
        _sand = XX_1.dot(Xres_2X).dot(XX_1)
        return _sand

    def variance(self, cov_type='nonrobust'):
        """Compute the covariance matrix for the fitted model.

        :param cov_type: Type of the covariance matrix, defaults to nonrobust.
        :type cov_type: :obj:`str`, optional

        :formulae: * **Non-robust covariance matrix**:

            .. math:: \\sigma_{\\beta} = {(X'SX)}^{-1}

            * **Sandwich covariance matrix**:

            .. math:: \\sigma_{\\beta} = {(X'X)}^{-1} (X'\\hat{S}X) {(X'X)}^{-1}

            with :math:`S = \\text{diag}(\\hat{p}_i(1 - \\hat{p}_i))` and
            :math:`\\hat{S} = \\text{diag} \\left( (Y_i - \\hat{p}_i)^{2} \\right)`

        .. note:: Only non-robust covariance matrix is currently available.
            Sandwich estimate and :math:`HC0`, :math:`HC1`, :math:`HC2`, :math:`HC3` will soon be implemented.

        :return: Fisher information matrix
        :rtype: :obj:`numpy.array`
        """
        self.cov_type = cov_type

        if self.cov_type == 'nonrobust':
            # Uses Fisher information matrix
            _variance = np.linalg.inv(self._hessian())
        # elif self.cov_type == 'sandwich':
        #     _variance = self._sandwich()
        else:
            error_msg = f"""Value for cov_type not valid. Got '{self.cov_type}'.
            Please chose 'nonrobust'.
            """
            raise ValueError(error_msg)
        return np.array(_variance)

    def _std_errors(self):
        """Compute the covariance matrix used for standard errors
        """
        _variance = self.variance()
        _std = np.sqrt((1 / self.n) * np.diag(_variance))
        return np.array(_std).reshape(self.p, 1)

    def fit(self, maxit=15, cov_type='nonrobust', improvement_threshold=0.0005, keep_hist=True, plot=False):
        """Fits the GLM regression model using Newton-Raphson method.

        :param maxit: Maximum number of iterations, defaults to 15.
        :type maxit: :obj:`int`, optional
        :param cov_type: Type of the covariance matrix (non-robust or sandwich), defaults to nonrobust.
        :type cov_type: :obj:`str`, optional
        :param improvement_threshold: Threshold from which we consider the likelihood improved, defaults to 0.0005.
        :type improvement_threshold: :obj:`float`, optional
        :param keep_hist: Keeps training history (gradients, hessian, etc...), can be turned off for saving memory, defaults to True.
        :type keep_hist: :obj:`bool`, optional
        :param plot: Plots evolution of log-likelihood through the different iterations (requires :obj:`keep_hist = True`), defaults to False.
        :type plot: :obj:`bool`, optional

        :formulae: * **Log-likelihood**:

            .. math:: l(\\beta) = y_{i} \\log \\left[ G(\\mathbf{x_i} \\beta) \\right] + (1 - y_{i}) \\log \\left[1 - G(\\mathbf{x_i} \\beta) \\right]

            * **Newton's method**:

            .. math:: \\hat{\\beta}_{s+1} = \\hat{\\beta}_{s} - H^{-1}_{s} G_{s}

            with

                .. math:: G_{s} = \\dfrac{\\partial}{\\partial \\beta_{s}} l(\\beta) = \\sum_{i=1}^{N} x_i (y_{i} - G(x_{i} \\beta))

            and

                .. math:: H_{s} = \\dfrac{\\partial^2}{\\partial \\beta_{s}^2} l(\\beta) = X'S_{s}X

            where :math:`S = \\text{diag} \\left( (Y_i - \\hat{p}_i)^{2} \\right)` and
            :math:`G` denotes the link function (:py:meth:`statinf.ml.activations.sigmoid` for logit or gaussian c.d.f for probit).

        :references: * Wooldridge, J. M. (2010). Econometric analysis of cross section and panel data.
            * Cameron, A. C., & Trivedi, P. K. (2005). Microeconometrics: methods and applications. Cambridge university press.
            * McCullagh, P. (2018). Generalized linear models. Routledge.
        """
        self.cov_type = cov_type.lower()
        self.it = 0
        diff = np.inf
        old_log_like = np.inf

        while (self.it < maxit) and (np.abs(diff) >= improvement_threshold):
            self.it      += 1
            log_like      = self._log_likelihood()
            gradient      = self._gradient()
            variance      = self.variance(cov_type=cov_type.lower())
            self.beta    -= variance.dot(gradient)
            diff          = old_log_like - log_like
            old_log_like  = log_like

            if keep_hist:
                self.log_likelihood = log_like
                self.log_likelihood_hist.append(log_like)
                self.gradient_hist.append(gradient)
                self.variance_hist.append(variance)

        # Check whether convergence was reached
        self.converged = np.abs(diff) < improvement_threshold
        if (self.converged is False) or (math.isnan(log_like)):
            warnings.warn('Model did not converge.', ConvergenceWarning)

        if plot:
            # Plot the training and test loss
            plt.title('GLM training history', loc='center')
            plt.plot(self.log_likelihood_hist, label='Log-likelihood')
            # plt.plot(self.test_losses, label='Test loss')
            plt.legend()
            plt.show()

    def r_squared(self):
        """Mc Fadden's pseudo-:math:`R^{2}` -- Goodness of fit

        :formula: .. math:: R^{2} = 1 - \\dfrac{LL(\\hat{\\beta})}{LL(\\bar{Y})}

        :return: Goodness of fit.
        :rtype: :obj:`float`
        """
        null_mod = self.__class__(formula=f'{self.Y_col} ~ 1', data=self.df)
        _r2 = 1 - self._log_likelihood() / null_mod._log_likelihood()
        self.r_squared = _r2
        return _r2

    def adjusted_r_squared(self):
        """Mc Fadden's pseudo-:math:`R^{2}` adjusted -- Adjusted goodness of fit

        :formula: .. math:: R^{2}_{adj} = 1 - \\dfrac{LL(\\hat{\\beta}) - p}{LL(\\bar{Y})}

        :return: Adjusted goodness of fit.
        :rtype: :obj:`float`
        """
        null_mod = self.__class__(formula=f'{self.Y_col} ~ 1', data=self.df)
        ar2 = 1 - (self._log_likelihood() - self.p) / null_mod._log_likelihood()
        self.adjusted_r_squared = ar2
        return ar2

    def summary(self, return_df=False):
        """Statistical summary for GLM model

        :param return_df: Return the summary as a Pandas DataFrame, else returns a string, defaults to False.
        :type return_df: :obj:`bool`

        :formulae: * **Mc Fadden's** :math:`R^{2}`:

            .. math:: R^{2} = 1 - \\dfrac{LL(\\hat{\\beta})}{LL(\\bar{Y})}

            where :math:`LL` represents the log-likelihood.

        :references: * Student. (1908). The probable error of a mean. Biometrika, 1-25.
            * Wooldridge, J. M. (2016). `Introductory econometrics: A modern approach
              <https://faculty.arts.ubc.ca/nfortin/econ495/IntroductoryEconometrics_AModernApproach_FourthEdition_Jeffrey_Wooldridge.pdf>`_.
              Nelson Education.
            * Agresti, A. (2003). Categorical data analysis (Vol. 482). John Wiley & Sons.
            * Murphy, K. P. (2012). Machine learning: a probabilistic perspective. MIT press.

        :return: Model's summary.
        :rtype: :obj:`pandas.DataFrame` or :obj:`str`
        """
        # Fit model if not already done
        if self.log_likelihood == []:
            self.fit()

        # Initialize
        betas = self.beta
        X = self._get_X()

        t_values = betas / self._std_errors()

        p_values = [round(2 * (1 - scp.t.cdf(np.abs(i[0]), (len(X) - 1))), 5) for i in t_values]
        z = scp.norm.ppf(0.975)

        summary_df = pd.DataFrame()
        summary_df["Variables"] = ['(Intercept)'] + self.X_col if self.fit_intercept else self.X_col
        summary_df["Coefficients"] = betas
        summary_df["Standard Errors"] = self._std_errors()
        summary_df["t-values"] = t_values
        summary_df["Probabilities"] = p_values
        summary_df["Significance"] = summary_df["Probabilities"].map(lambda x: get_significance(x))
        summary_df["CI_lo"] = summary_df["Coefficients"] - z * summary_df["Standard Errors"]
        summary_df["CI_hi"] = summary_df["Coefficients"] + z * summary_df["Standard Errors"]

        # Null model for null log likelihood
        null_mod = self.__class__(formula=f'{self.Y_col} ~ 1', data=self.df)

        # Key metrics for summary
        _ll = round(self._log_likelihood(), 2)
        nll = round(null_mod._log_likelihood(), 2)
        _r2 = round(1 - _ll / nll, 5)
        ar2 = round(1 - (_ll - self.p) / nll, 5)
        _n_ = self.n
        _p_ = self.p
        _it = self.it
        _cv = '     True' if self.converged else "    False"
        _ct = ' ' + self.cov_type

        # Likelihood Ratio test
        G = 2 * (_ll - nll)
        lrp = round(scp.chi2.sf(G, self.p), 3)

        if return_df:
            return(summary_df)
        else:
            max_var = np.max([len(v) for v in summary_df.Variables])

            add_sp = ' ' * np.max([max_var - 17, 0])
            add_sep = '=' * np.max([max_var - 17, 0])
            space = np.max([max_var, 17])

            summ = f"============================================================================================================={add_sep}\n"
            summ += f'|                                              Logit summary                                                {add_sp}|\n'
            summ += f"============================================================================================================={add_sep}\n"
            summ += f"| McFadden's R²          =              {_r2:10} | McFadden's R² Adj.         =                {ar2:10} {add_sp}|\n"
            summ += f"| Log-Likelihood         =              {_ll:10} | Null Log-Likelihood        =                {nll:10} {add_sp}|\n"
            summ += f"| LR test p-value        =              {lrp:10} | Covariance                 =                {_ct:10} {add_sp}|\n"
            summ += f"| n                      =              {_n_:10} | p                          =                {_p_:10} {add_sp}|\n"
            summ += f"| Iterations             =              {_it:10} | Convergence                =                 {_cv:5} {add_sp}|\n"
            summ += summary(summary_df)
            return(summ)

    def predict(self, new_data, return_proba=False):
        """Predicted :math:`\\hat{Y}` values for for a new dataset

        :param new_data: New data to evaluate with pandas data-frame format.
        :type new_data: :obj:`pandas.DataFrame`
        :param return_proba: Whether to return probabilities or binary values, defaults to False
        :type return_proba: :obj:`bool`, optional

        :formula: .. math:: f(X) = \\dfrac{1}{e^{-\\beta X} + 1}

        :return: Predictions :math:`\\hat{Y}`
        :rtype: :obj:`numpy.ndarray`
        """
        # Fit model if not already done
        if self.log_likelihood == []:
            self.fit()

        # Tranform data
        new_X = self._get_X(new_data=new_data)

        pred = self._prob(new_X)
        if return_proba:
            return np.asarray([x[0] for x in pred])
        else:
            return np.asarray([1 if x[0] > 0.5 else 0 for x in pred])

    def partial_effects(self, variables, new_data=None, average=False):
        """Computes Partial and Average Partial Effects (APE).

        :param variables: List of variables for which to compute the PE/APE.
        :type variables: :obj:`list`
        :param new_data: Data to use for computations, optional, defaults to None (uses training set).
        :type new_data: :obj:`pandas.DataFrame`
        :param average: Whether to compute Average Partial Effects or not, defaults to False.
        :type average: :obj:`bool`, optional

        :formula: .. math:: PE(X_{i}) = \\beta_{i} \\dfrac{e^{-\\beta X}}{(e^{-\\beta X} + 1)^{2}}

        if :obj:`average = True`, :math:`APE = \\bar{PE}`

        :raises TypeError: If argument :obj:`variables` is neither :obj:`str` nor :obj:`list`.

        :return: Dictionnary including Partial Effects (PE) or Average Partial Effects (APE).
        :rtype: :obj:`dict`
        """

        assert self.family == 'binomial', 'APE module only available for Logit model.'

        # Fit model if not already done
        if self.log_likelihood == []:
            self.fit()

        # Get data
        X = self._get_X(new_data)

        # Get coefficients as a DataFrame (easier to handle for fitlering)
        if self.fit_intercept:
            # Append the name of the intercept
            cols = ['(Intercept)'] + self.X_col
        else:
            cols = self.X_col

        coefs = pd.DataFrame({'Coefficients': [x[0] for x in self.beta]}, index=cols)

        # Compute exp(bX)
        exp_bx = np.exp(-X.dot(self.beta))

        # Check type for argument variables
        if type(variables) == str:
            var = [variables]
        elif type(variables) == list:
            var = variables
        else:
            raise TypeError('Type for argument variables is not valid.')

        # Initialize the dictionnary for output
        pes = {}

        # Compute PEs (or APEs) for each variable
        for v in var:
            b = coefs.Coefficients[v].min()
            pe = b * exp_bx / ((exp_bx + 1)**2)
            pe_out = pe.mean() if average else pe

            pes.update({v: pe_out})

        return pes



import numpy as np
import pandas as pd

from scipy.stats import norm

from ..misc import test_summary, format_object
from ..regressions.LinearModels import OLS


# This is being validated with statsmodels implementation:
# https://github.com/statsmodels/statsmodels/blob/master/statsmodels/tsa/stattools.py#L159
# As well as checked with the litterature cited below
"""
Aligned with the mackinnonp function from adfvalues in statsmodels
"""
_stats_value_max = {
    "none": [-1.04, -1.53, -2.68, -3.09, -3.07, -3.77],
    "c": [-1.61, -2.62, -3.13, -3.47, -3.78, -3.93],
    "ct": [-2.89, -3.19, -3.50, -3.65, -3.80, -4.36],
    # "ctt": [-3.21, -3.51, -3.81, -3.83, -4.12, -4.63],
}

_dist_quant_small = {
    "none": [0.6344, 1.2378, 3.2496 * 1e-2],
    "c": [2.1659, 1.4412, 3.8269 * 1e-2],
    "ct": [3.2512, 1.6047, 4.9588 * 1e-2],
    "ctt": [4.0003, 1.658, 4.8288 * 1e-2],
}

_dist_quant_large = {
    "none": [0.4797, 9.3557, -0.6999, 3.3066],
    "c": [1.7339, 9.3202, -1.2745, -1.0368],
    "ct": [2.5261, 6.1654, -3.7956, -6.0285],
    "ctt": [3.0778, 4.9529, -4.1477, -5.9359],
}

_crit_values = {
    'none': [-2.56574, -1.941, -1.61682],
    'c': [-3.43035, -2.86154, -2.56677],
    'ct': [-3.95877, -3.41049, -3.12705],
    'ctt': [-4.37113, -3.83239, -3.55326],
}

# MacKinnon p-values
def _MacKinnon_pvalues(statvalue, trend, k=1):
    if statvalue <= _stats_value_max[trend][k - 1]:
        values = _dist_quant_small[trend]
    else:
        values = np.array(_dist_quant_large[trend]) * np.array([1, 1e-1, 1e-1, 1e-2])

    return norm.cdf(np.polyval(values[::-1], statvalue))


# Augmented Dickey-Fuller
def adf_test(x, lag='auto', trend='c', metric='aic', return_tuple=False):
    """
    Augmented Dickey-Fuller test.

    The test is primarily used for unit root testing for univariate time series in order to check for stationnarity.
    The null hypothesis :math:`H_0` is that the root is 1 and series is not stationary.

    The methodology checks the T-statistic of the variable in the linear regression and compares with critical value from MacKinnon.
    The default behavior will determine the lag minimizing the information criteria (AIC or BIC). User can also choose the lag.


    :param x: Time series to be tested.
    :type x: :obj:`pandas.DataFrame` or :obj:`numpy.ndarray`
    :param lag: Lag to be considered in the series, defaults to 'auto'.
    :type lag: :obj:`int`, optional
    :param trend: Trend to be included in the regression, defaults to 'c'. Values can be:

        * 'none': only consider the lag.
        * 'c': only include a constant.
        * 't': only include the trend component.
        * 'ct': include both constant and trend components.

    :type trend: :obj:`str`, optional
    :param metric: Type of metric to use in the linear regression for selecting the optimal lag, defaults to 'aic'. Can either be 'aic' or 'bic'.
    :type metric: :obj:`str`, optional
    :param return_tuple: Return a tuple with test statistic, p-value, lag. Defaults to False.
    :type return_tuple: :obj:`bool`

    :example:

    >>> from statinf import stats
    >>> stats.adf_test(series,  trend='ct')
    ... +------------------------------------------------------------+
    ... |                     Dickey-Fuller test                     |
    ... +------------+----------------+------------+---------+-------+
    ... |     df     | Critical value | Stat value | p-value |   H0  |
    ... +------------+----------------+------------+---------+-------+
    ... |        107 |       -3.41049 |  -1.298438 | 0.88831 | True  |
    ... +------------+----------------+------------+---------+-------+
    ... * We cannot reject H0: the series is not stationarity
    ... * Used 12 lags


    :return: Summary for the test or tuple statistic, critical value, p-value.
    :rtype: :obj:`str` or :obj:`tuple`


    :references: * Hamilton, J. D. (1994). Time series analysis. Princeton university press, Princeton, NJ.
        * Wooldridge, J. M. (2010). Econometric analysis of cross section and panel data. MIT press.
        * Cameron, A. C., & Trivedi, P. K. (2009). Microeconometrics using stata (Vol. 5, p. 706). College Station, TX: Stata press.
        * MacKinnon, J. G. (2010). Critical values for cointegration tests (No. 1227). Queen's Economics Department Working Paper.
    """

    assert trend in ('none', 'c', 'ct'), f"The value for trend needs to be 'none', 'c' or 'ct'. Got '{trend}'"

    _df = pd.DataFrame(x).reset_index(drop=True)
    _df.columns = ['x']
    _df['trend'] = _df.index + 1.
    n = _df.shape[0]

    nb_ext = {'none': 0, 'c': 0, 't': 1, 'ct': 2}

    maxlag = int(np.ceil(12.0 * np.power(n / 100.0, 1 / 4.0)))
    # -1 for the diff
    maxlag = min(n // 2 - nb_ext[trend] - 1, maxlag)

    if lag != 'auto':
        assert lag <= maxlag, ValueError(f'Lag value is too high. Maximum is {maxlag}, got {lag}')

    for _lag in range(1, maxlag + 2):
        _df[f'x_lag_{_lag}'] = _df.x.shift(_lag)

    for _lag in range(1, maxlag + 2):
        if _lag == 1:
            _df['delta_lag_1'] = _df['x'] - _df['x_lag_1']
        else:
            _df[f'delta_lag_{_lag}'] = _df[f'x_lag_{_lag - 1}'] - _df[f'x_lag_{_lag}']

    filtered = _df.dropna().copy()

    # Define whether we need to add trend component
    trend_var = ' + trend ' if trend in ['t', 'ct'] else ''
    # Add an intercept if user want a trend
    cst = trend in ['c', 'ct']

    base_cols = ['delta_lag_1', 'x_lag_1', 'trend']

    # Defining lag
    if lag == 'auto':
        perfs = {}
        # Define optimal lag
        for _lag in range(1, maxlag + 1):
            _cols = [f'delta_lag_{i}' for i in range(2, _lag + 2)]
            _lag_cols = ' + '.join(_cols)
            _reg_formula = 'delta_lag_1 ~ x_lag_1 + ' + _lag_cols
            # Add constant variable if user requests
            _reg_formula += trend_var
            _temp_df = _df[_cols + base_cols].dropna().copy()
            # Fit OLS
            temp_ols = OLS(_reg_formula, data=_temp_df, fit_intercept=cst)
            # perfs[l] = temp_ols
            perfs.update({temp_ols._aic(): _lag})

        best_perf = np.min([k for k in perfs.keys()])
        best_lag = perfs.get(best_perf) - 1
    else:
        best_lag = lag

    # Parse lagged columns to use as formula
    cols_lag = [f'delta_lag_{i}' for i in range(2, best_lag + 2)]

    if best_lag == 0:
        _reg_formula = 'delta_lag_1 ~ x_lag_1'
    else:
        _lag_cols = ' + '.join(cols_lag)
        _reg_formula = 'delta_lag_1 ~ x_lag_1 + ' + _lag_cols

    # Add constant variable if user requests
    _reg_formula += trend_var

    # Prepare final data set with created lags
    filtered = _df[cols_lag + base_cols].dropna().copy()
    n = filtered.shape[0]

    # Run OLS and get the stat value of x_lag_1 from summary
    _ols = OLS(_reg_formula, data=filtered, fit_intercept=cst)
    _summ = _ols.summary(True)
    stat_value = _summ['t-values'][_summ.Variables == 'x_lag_1'].max()

    # Get the MacKinnon pvalue based on the OLS stat value
    p_val = _MacKinnon_pvalues(statvalue=stat_value, trend=trend, k=1)

    # Format summary output
    _summ = test_summary(df=n, critical_value=_crit_values[trend][1], t_value=stat_value,
                         p_value=p_val, title='Dickey-Fuller test',
                         extra=f' * Used {best_lag} lags',
                         h0='the series is not stationarity',
                         h1='the series is stationarity')

    if return_tuple:
        return stat_value, p_val, best_lag, n
    else:
        return _summ


# Cointegration test
def coint_test(x1, x2, lag='auto', trend='c', metric='aic', return_tuple=False):
    """
    Cointegration test.

    Checks the relation between two time series. The null hypothesis :math:`H_0` is that there is no cointegration between both series.
    The methodology consists in regressing one variable on the other and checks the stationarity of the residuals.

    :param x1: First series to be compared.
    :type x1: :obj:`pandas.DataFrame` or :obj:`numpy.ndarray`
    :param x2: Second series to be compared.
    :type x2: :obj:`pandas.DataFrame` or :obj:`numpy.ndarray`
    :param lag: Lag to be considered in the series, defaults to 'auto'.
    :type lag: :obj:`int`, optional
    :param trend: Trend to be included in the regression, defaults to 'c'. See details in :py:meth:`statinf.stats.timeseries.adf_test`.
    :type trend: :obj:`str`, optional
    :param metric: Type of metric to use in the linear regression for selecting the optimal lag, defaults to 'aic'. Can either be 'aic' or 'bic'.
    :type metric: :obj:`str`, optional
    :param return_tuple: [description], defaults to False
    :param return_tuple: Return a tuple with test statistic, p-value, and relation. Defaults to False.
    :type return_tuple: :obj:`bool`

    :example:

    >>> from statinf import stats
    >>> stats.coint_test(series, series2,  trend='ct')
    ... +------------------------------------------------------------+
    ... |                     Cointegration test                     |
    ... +------------+----------------+------------+---------+-------+
    ... |     df     | Critical value | Stat value | p-value |   H0  |
    ... +------------+----------------+------------+---------+-------+
    ... |        120 |       -3.41049 |  -2.213066 |  0.4825 | True  |
    ... +------------+----------------+------------+---------+-------+
    ... * We cannot reject H0: the series are not cointegrated

    :return: Summary for the test or tuple statistic, critical value, p-value.
    :rtype: :obj:`str` or :obj:`tuple`

    :references: * Hamilton, J. D. (1994). Time series analysis. Princeton university press, Princeton, NJ.
        * Wooldridge, J. M. (2010). Econometric analysis of cross section and panel data. MIT press.
        * MacKinnon, J. G. (1991). Critical values for cointegration tests. In Eds., Long-Run Economic Relationship: Readings in Cointegration.
    """
    _df = pd.DataFrame(x1).reset_index(drop=True)
    _df.columns = ['x1']
    _df['x2'] = format_object(x2, name='x2')
    _df['trend'] = _df.index + 1.
    _df['cst'] = 1.
    n, k = _df.shape

    # Add an intercept if user want a trend
    cst = trend in ['c', 'ct']
    _reg_formula = 'x1 ~ x2'
    # Define whether we need to add trend component
    _reg_formula += ' + trend ' if trend in ['t', 'ct'] else ''

    _ols = OLS(_reg_formula, data=_df, fit_intercept=cst)
    _summ_ols = _ols.summary(True)
    res = _ols._get_error()
    # sm_ols = lm.OLS(x1, _df[['x2', 'trend', 'cst']]).fit()
    # print(res)
    # print(sm_ols.resid)

    if _ols.r_squared() < 1. - 1e-4:
        # In case of no perfect fit
        stat_value, p, bl, _n = adf_test(res, lag=lag, trend='none', metric=metric, return_tuple=True)
    else:
        raise ValueError('Your time series seem to be colinear')

    p_val = _MacKinnon_pvalues(statvalue=stat_value, trend=trend, k=k)

    vars_rel = 'negative' if _summ_ols.Coefficients[(_summ_ols.Variables == 'x2')].max() < 0 else 'positive'

    _summ = test_summary(df=n, critical_value=_crit_values[trend][1], t_value=stat_value,
                         p_value=p_val, title='Cointegration test',
                         h1_conclu=f' * The series seem to have a {vars_rel} relation',
                         h0='the series are not cointegrated',
                         h1='the series are cointegrated')

    if return_tuple:
        return stat_value, p_val, vars_rel
    else:
        return _summ



import numpy as np
import pandas as pd
import re
import warnings
from types import SimpleNamespace


# Ranking data

def rankdata(x):
    """Assigns rank to data.
    This is mainly used for analysis like Spearman's correlation.

    :param x: Input vector. Format can be :obj:`numpy.array`, :obj:`list` or :obj:`pandas.Series`.
    :type x: :obj:`numpy.array`

    :example:

    >>> rankdata([2., 5.44, 3.93, 3.3, 1.1])
    ... array([1, 4, 3, 2, 0])

    :return: Vector with ranked values.
    :rtype: :obj:`numpy.array`
    """
    x_arr = np.asarray(x)
    sorted_array = sorted(x_arr)
    rk = [sorted_array.index(i) for i in x_arr]
    return np.array(rk)


#######################################################################################################################

# Parse formula and data transformations

def parse_formula(formula, data, check_values=True, return_all=False):
    """This function is used in regression models in order to apply transformations on the data from a formula.
    It allows to apply transformations from a :obj:`str` formula. See below for examples.

    :param formula: Regression formula to be run of the form :obj:`y ~ x1 + x2`. Accepted functions are:

        * :math:`\\log(x)` \\: :obj:`log(X)`
        * :math:`\\exp(x)` \\: :obj:`exp(X)`
        * :math:`\\sqrt{x}` \\: :obj:`sqrt(X)`
        * :math:`\\cos(x)` \\: :obj:`cos(X)`
        * :math:`\\sin(x)` \\: :obj:`sin(X)`
        * :math:`x^{z}` \\: :obj:`X ** Z`
        * :math:`\\dfrac{x}{z}` \\: :obj:`X/Z`
        * :math:`x \\times z` \\: :obj:`X*Z`

    :type formula: :obj:`str`
    :param data: Data on which to perform the transformations.
    :type data: :obj:`pandas.DataFrame`
    :param check_values: For each transformation check whether the data range satisfy the domain definition of the function, defaults to True.
    :type check_values: bool, optional
    :param return_all: Returns the transformed data, column :obj:`Y` and columns :obj:`X`, defaults to False.
    :type return_all: bool, optional

    :example:

    >>> from statinf.data import parse_formula
    >>> print(input_df)
    ... +-----------+-----------+-----------+
    ... |        X1 |        X2 |         Y |
    ... +-----------+-----------+-----------+
    ... |  0.555096 |  0.681083 | -1.383428 |
    ... |  1.155661 |  0.391129 | -7.780989 |
    ... | -0.299251 | -0.445602 | -8.146673 |
    ... | -0.978311 |  1.312146 |  8.653818 |
    ... | -0.225917 |  0.522016 | -9.684332 |
    ... +-----------+-----------+-----------+
    >>> form = 'Y ~ X1 + X2 + exp(X2) + X1*X2'
    >>> new_df = parse_formula(form, data=input_df)
    >>> print(new_df)
    ... +-----------+-----------+-----------+-----------+-----------+
    ... |        X1 |        X2 |         Y |   exp(X2) |     X1*X2 |
    ... +-----------+-----------+-----------+-----------+-----------+
    ... |  0.555096 |  0.681083 | -1.383428 |  1.976017 |  0.378066 |
    ... |  1.155661 |  0.391129 | -7.780989 |  1.478649 |  0.452012 |
    ... | -0.299251 | -0.445602 | -8.146673 |  0.640438 |  0.133347 |
    ... | -0.978311 |  1.312146 |  8.653818 |  3.714134 | -1.283687 |
    ... | -0.225917 |  0.522016 | -9.684332 |  1.685422 | -0.117932 |
    ... +-----------+-----------+-----------+-----------+-----------+

    :raises ValueError: Returns an error when the data cannot satisfy the domain definition for the required transformation.

    :return: Transformed data set
    :rtype: :obj:`pandas.DataFrame`
    """
    warnings.filterwarnings('ignore')
    # Parse formula
    no_space_formula = formula.replace(' ', '')
    Y_col = no_space_formula.split('~')[0]
    X_col = no_space_formula.split('~')[1].split('+')

    # Non-linear transformations
    log_cols = [re.search('(?<=log\\().*?(?=\\))', x).group(0) for x in X_col if re.findall('log\\(', x)]  # log
    exp_cols = [re.search('(?<=exp\\().*?(?=\\))', x).group(0) for x in X_col if re.findall('exp\\(', x)]  # exp
    sqrt_cols = [re.search('(?<=sqrt\\().*?(?=\\))', x).group(0) for x in X_col if re.findall('sqrt\\(', x)]  # sqrt
    cos_cols = [re.search('(?<=cos\\().*?(?=\\))', x).group(0) for x in X_col if re.findall('cos\\(', x)]  # cos
    sin_cols = [re.search('(?<=sin\\().*?(?=\\))', x).group(0) for x in X_col if re.findall('sin\\(', x)]  # sin

    # Transformation functions
    transformations_functional = {'log': {'func': np.log, 'cols': log_cols},
                                  'exp': {'func': np.exp, 'cols': exp_cols},
                                  'cos': {'func': np.cos, 'cols': cos_cols},
                                  'sin': {'func': np.sin, 'cols': sin_cols},
                                  'sqrt': {'func': np.sqrt, 'cols': sqrt_cols},
                                  }
    # Apply transformations
    for key, transformation in transformations_functional.items():
        for c in transformation['cols']:
            col_to_transform = c  # .split('(')[1].split(')')[0]
            # Transform
            data.loc[:, f'{key}({col_to_transform})'] = transformation['func'](data[col_to_transform])

    # Multiplications, power and ration functions
    pow_cols = [x for x in X_col if re.findall('[a-zA-Z0-9\\(\\)][*][*][a-zA-Z0-9]', x)]  # X1 ** x
    inter_cols = [x for x in X_col if re.findall('[a-zA-Z0-9\\(\\)][*][a-zA-Z0-9]', x)]  # X1 * X2
    div_cols = [x for x in X_col if re.findall('[a-zA-Z0-9\\(\\)][/][a-zA-Z0-9]', x)]  # X1 / X2

    # Exponents
    for c in pow_cols:
        c_left = c.split('**')[0]
        c_power = c.split('**')[1]
        # Get components as number or column from data
        left = data[c_left].values if c_left in data.columns else float(c_left)
        power = data[c_power].values if c_power in data.columns else float(c_power)
        # Transform
        data.loc[:, c] = left ** power
    # Multiplications
    for c in inter_cols:
        c_left = c.split('*')[0]
        c_right = c.split('*')[1]
        # Get components as number or column from data
        try:
            left = data[c_left].values if c_left in list(data.columns) + X_col else float(c_left)
            right = data[c_right].values if c_right in list(data.columns) + X_col else float(c_right)
        except Exception:
            raise ValueError(f'Columns {c_left} or {c_right} not found in data.')
        # Transform
        data.loc[:, c] = left * right
    # Divide
    for c in div_cols:
        c_num = c.split('/')[0]
        c_denom = c.split('/')[1]
        # Get components as number or column from data
        num = data[c_num].values if c_num in list(data.columns) + X_col else float(c_num)
        denom = data[c_denom].values if c_denom in list(data.columns) + X_col else float(c_denom)
        if check_values:
            assert (denom == 0.).sum() == 0, f'Column {col_to_transform} contains null values.'
        # Transform
        data.loc[:, c] = num / denom
    if '1' in X_col:
        data['1'] = 1

    # Putting pandas' warning message back
    warnings.filterwarnings('default')

    if return_all:
        return data, X_col, Y_col
    else:
        return data


#######################################################################################################################

# Adding One Hot Encoding
def OneHotEncoding(data, columns, drop=True, verbose=False):
    """Performs One Hot Encoding (OHE) usally used in Machine Learning.

    :param data: Data Frame on which we apply One Hot Encoding.
    :type data: :obj:`pandas.DataFrame`
    :param columns: Column to be converted to dummy variables.
    :type columns: :obj:`list`
    :param drop: Drop the column that needs to be converted to dummies, defaults to True.
    :type drop: :obj:`bool`, optional
    :param verbose: Display progression, defaults to False.
    :type verbose: :obj:`bool`, optional

    :example:

        >>> from statinf.data import OneHotEncoding
        >>> print(df)
        ... +----+--------+----------+-----+
        ... | Id | Gender | Category | Age |
        ... +----+--------+----------+-----+
        ... |  1 | Male   |        A |  23 |
        ... |  2 | Female |        B |  21 |
        ... |  3 | Female |        A |  31 |
        ... |  4 | Male   |        C |  22 |
        ... |  5 | Female |        A |  26 |
        ... +----+--------+----------+-----+
        >>> # Encoding columns "Gender" and "Category"
        >>> new_df = OneHotEncoding(df, columns=["Gender", "Category"])
        >>> print(new_df)
        ... +----+---------------+------------+------------+-----+
        ... | Id | Gender_Female | Category_A | Category_B | Age |
        ... +----+---------------+------------+------------+-----+
        ... |  1 |             0 |          1 |          0 |  23 |
        ... |  2 |             1 |          0 |          1 |  21 |
        ... |  3 |             1 |          1 |          0 |  31 |
        ... |  4 |             0 |          0 |          0 |  22 |
        ... |  5 |             1 |          1 |          0 |  26 |
        ... +----+---------------+------------+------------+-----+

    :return: Transformed data with One Hot Encoded variables.
    :rtype: :obj:`pandas.DataFrame`
    """

    dataset = data.copy()

    try:
        if dataset.meta._ohe_exists:
            dataset.meta._ohe_exists = True
    except Exception:
        dataset.meta = SimpleNamespace()
        dataset.meta._ohe_exists = True
        dataset.meta._ohe = {}

    cols = [columns] if type(columns) == str else columns

    # Start encoding column by column
    for column in cols:
        # Get all values from the column
        all_values = dataset[column].unique()
        # Add column metadata
        dataset.meta._ohe.update({column: [f'{column}_{val}' for val in all_values]})

        # Encode values
        for val in all_values:
            if verbose:
                print('Encoding for value: ' + str(val))
            colname = column + '_' + str(val)
            dataset.loc[:, colname] = 0
            dataset.loc[dataset[column] == val, colname] = 1

        if drop:
            dataset.drop(columns=[column], inplace=True)

    return(dataset)


#######################################################################################################################

# Convert an array of values into a dataset matrix: used for LSTM data pre-processing
def create_dataset(data, n_in=1, n_out=1, dropnan=True):
    """Function to convert a DataFrame into into multivariate time series format readable by Keras LSTM.

    :param data: DataFrame on which to aply the transformation.
    :type data: :obj:`pandas.DataFrame`
    :param n_in: Input dimension also known as look back or size of the window, defaults to 1
    :type n_in: :obj:`int`, optional
    :param n_out: Output dimension, defaults to 1
    :type n_out: :obj:`int`, optional
    :param dropnan: Remove empty values in the series, defaults to True
    :type dropnan: :obj:`bool`, optional

    :return: Features converted for Keras LSTM.
    :rtype: :obj:`pandas.DataFrame`
    """
    n_vars = 1 if type(data) is list else data.shape[1]
    df = pd.DataFrame(data)
    cols, names = list(), list()
    # input sequence (t-n, ... t-1)
    for i in range(n_in, 0, -1):
        cols.append(df.shift(i))
        names += [('var%d(t-%d)' % (j + 1, i)) for j in range(n_vars)]
    # forecast sequence (t, t+1, ... t+n)
    for i in range(0, n_out):
        cols.append(df.shift(-i))
        if i == 0:
            names += [('var%d(t)' % (j + 1)) for j in range(n_vars)]
        else:
            names += [('var%d(t+%d)' % (j + 1, i)) for j in range(n_vars)]
    # put it all together
    agg = pd.concat(cols, axis=1)
    agg.columns = names
    # drop rows with NaN values
    if dropnan:
        agg.dropna(inplace=True)
    return agg


def split_sequences(data, look_back=1):
    """Split a multivariate time series from :py:meth:`statinf.data.ProcessData.multivariate_time_series` into a Keras' friendly format for LSTM.

    :param data: Data in the format of sequences to transform.
    :type data: :obj:`numpy.ndarray`
    :param look_back: Size of the trailing window, number of time steps to consider, defaults to 1.
    :type look_back: :obj:`int`

    :exemple:

        >>> from statinf.data import multivariate_time_series, split_sequences
        >>> train_to_split = multivariate_time_series(train)
        >>> X, y = split_sequences(train_to_split, look_back=7)

    :return: * :obj:`x`: Input data converted for Keras LSTM.
        * :obj:`y`: Target series converted for Keras LSTM.
    :rtype: * :obj:`numpy.ndarray`
        * :obj:`numpy.ndarray`
    """
    X, y = list(), list()
    for i in range(len(data)):
        # find the end of this pattern
        end_ix = i + look_back
        # check if we are beyond the dataset
        if end_ix > len(data) - 1:
            break
        # gather input and output parts of the pattern
        seq_x, seq_y = data[i:end_ix, :], data[end_ix, :]
        X.append(seq_x)
        y.append(seq_y)
    return np.array(X), np.array(y)


def multivariate_time_series(data):
    """Convert a dataframe into numpy array multivariate time series.

    :param data: Input data to transform.
    :type data: :obj:`pandas.DataFrame`

    :exemple:

        >>> from statinf.data import multivariate_time_series, split_sequences
        >>> train_to_split = multivariate_time_series(train)
        >>> X, y = split_sequences(train_to_split, look_back=7)

    :return: Transformed multivariate time series data.
    :rtype: :obj:`numpy.ndarray`
    """
    to_stack = ()
    for _c in data.columns:
        series = data[_c].to_numpy()
        to_stack += (series.reshape((len(series), 1)), )
    return np.hstack(to_stack)


#######################################################################################################################

# Scale dataset
class Scaler:
    def __init__(self, data, columns):
        """Data scaler.

        :param data: Data set to scale.
        :type data: :obj:`pandas.DataFrame`
        :param columns: Columns to scale.
        :type columns: :obj:`list`

        :example:

            >>> from statinf.data import Scaler, generate_dataset
            >>> coeffs = [1.2556, -0.465, 1.665414, 2.5444, -7.56445]
            >>> data = generate_dataset(coeffs, n=10, std_dev=2.6)
            >>> # Original dataset
            >>> print(data)
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            ... |        X0 |        X1 |        X2 |        X3 |        X4 |         Y |
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            ... |  0.977594 |  1.669510 | -1.385569 |  0.696975 | -1.207098 |  8.501692 |
            ... | -0.953802 |  1.025392 | -0.639291 |  0.658251 |  0.746814 | -7.186085 |
            ... | -0.148140 | -0.972473 |  0.843746 |  1.306845 |  0.269834 |  1.939924 |
            ... |  0.499385 | -1.081926 |  2.646441 |  0.910503 |  0.857189 |  0.389257 |
            ... | -0.563977 | -0.511933 | -0.726744 | -0.630345 | -0.486822 | -0.125787 |
            ... | -0.434994 | -0.396210 |  1.101739 | -0.660236 | -1.197566 |  7.735832 |
            ... |  0.032478 | -0.114952 | -0.097337 |  1.794769 |  1.239423 | -5.510332 |
            ... |  0.085569 | -0.600019 |  0.224186 |  0.301771 |  1.278387 | -8.648084 |
            ... | -0.028844 | -0.329940 | -0.301762 |  0.946077 | -0.359133 |  5.099971 |
            ... | -0.665312 |  0.270254 | -1.263288 |  0.545625 |  0.499162 | -6.126528 |
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            >>> # Load scaler class
            >>> scaler = Scaler(data=data, columns=['X1', 'X2'])
            >>> # Scale our dataset with MinMax method
            >>> scaled_df = scaler.MinMax()
            >>> print(scaled_df)
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            ... |        X0 |        X1 |        X2 |        X3 |        X4 |         Y |
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            ... |  0.977594 |  1.000000 |  0.000000 |  0.696975 | -1.207098 |  8.501692 |
            ... | -0.953802 |  0.765898 |  0.185088 |  0.658251 |  0.746814 | -7.186085 |
            ... | -0.148140 |  0.039781 |  0.552904 |  1.306845 |  0.269834 |  1.939924 |
            ... |  0.499385 |  0.000000 |  1.000000 |  0.910503 |  0.857189 |  0.389257 |
            ... | -0.563977 |  0.207162 |  0.163399 | -0.630345 | -0.486822 | -0.125787 |
            ... | -0.434994 |  0.249221 |  0.616890 | -0.660236 | -1.197566 |  7.735832 |
            ... |  0.032478 |  0.351444 |  0.319501 |  1.794769 |  1.239423 | -5.510332 |
            ... |  0.085569 |  0.175148 |  0.399244 |  0.301771 |  1.278387 | -8.648084 |
            ... | -0.028844 |  0.273307 |  0.268801 |  0.946077 | -0.359133 |  5.099971 |
            ... | -0.665312 |  0.491445 |  0.030328 |  0.545625 |  0.499162 | -6.126528 |
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            >>> # Unscale the new dataset to retreive previous data scale
            >>> unscaled_df = scaler.unscaleMinMax(scaled_df)
            >>> print(unscaled_df)
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            ... |        X0 |        X1 |        X2 |        X3 |        X4 |         Y |
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
            ... |  0.977594 |  1.669510 | -1.385569 |  0.696975 | -1.207098 |  8.501692 |
            ... | -0.953802 |  1.025392 | -0.639291 |  0.658251 |  0.746814 | -7.186085 |
            ... | -0.148140 | -0.972473 |  0.843746 |  1.306845 |  0.269834 |  1.939924 |
            ... |  0.499385 | -1.081926 |  2.646441 |  0.910503 |  0.857189 |  0.389257 |
            ... | -0.563977 | -0.511933 | -0.726744 | -0.630345 | -0.486822 | -0.125787 |
            ... | -0.434994 | -0.396210 |  1.101739 | -0.660236 | -1.197566 |  7.735832 |
            ... |  0.032478 | -0.114952 | -0.097337 |  1.794769 |  1.239423 | -5.510332 |
            ... |  0.085569 | -0.600019 |  0.224186 |  0.301771 |  1.278387 | -8.648084 |
            ... | -0.028844 | -0.329940 | -0.301762 |  0.946077 | -0.359133 |  5.099971 |
            ... | -0.665312 |  0.270254 | -1.263288 |  0.545625 |  0.499162 | -6.126528 |
            ... +-----------+-----------+-----------+-----------+-----------+-----------+
        """
        super(Scaler, self).__init__()
        self.data = data.copy()
        self.scalers = {}
        self.columns = list(columns)

        for c in columns:
            _min = self.data[c].min()
            _max = self.data[c].max()
            _mean = self.data[c].mean()
            _std = self.data[c].std()
            _scale_temp = {'min': float(_min),
                           'max': float(_max),
                           'mean': float(_mean),
                           'std': float(_std),
                           }
            self.scalers.update({c: _scale_temp})

    def _col_to_list(self, columns):
        """Transforms column names to be scaled as list.

        :param columns: Column names to be scaled.
        :type columns: :obj:`list` or :obj:`str`

        :return: Column name(s) as a list
        :rtype: :obj:`list`
        """
        if columns is None:
            cols = self.columns
        elif type(columns) == str:
            cols = [columns]
        else:
            cols = columns
        return cols

    def MinMax(self, data=None, columns=None, feature_range=(0, 1), col_suffix=''):
        """Min-max scaler. Data we range between 0 and 1.

        :param data: Data set to scale, defaults to None, takes data provided in :py:meth:`__init__`, defaults to None.
        :type data: :obj:`pandas.DataFrame`, optional
        :param columns: Columns to be scaled, defaults to None, takes the list provided in :py:meth:`__init__`, defaults to None.
        :type columns: :obj:`list`, optional
        :param feature_range: Expected value range of the scaled data, defaults to (0, 1).
        :type feature_range: :obj:`tuple`, optional
        :param col_suffix: Suffix to add to colum names, defaults to '', overrides the existing columns.
        :type col_suffix: :obj:`str`, optional

        :formula: .. math:: x_{\\text{scaled}} = \\dfrac{x - \\min(x)}{\\max(x) - \\min(x)} \\cdot (f\\_max - f\\_min) + f\\_min

            where :math:`(f\\_min, f\\_max)` defaults to :math:`(0, 1)` and corresponds to the expected data range of the scaled data from argument :obj:`feature_range`.

        :return: Data set with scaled features.
        :rtype: :obj:`pandas.DataFrame`
        """
        self._minmax_suffix = col_suffix
        self._minmax_feature_range = feature_range
        f_min, f_max = self._minmax_feature_range
        cols = self._col_to_list(columns)
        df = self.data if data is None else data.copy()

        for c in cols:
            # Retreive min and max
            _min = self.scalers[c]['min']
            _max = self.scalers[c]['max']
            tmp = (df[c] - _min) / (_max - _min)
            df[c + col_suffix] = tmp * (f_max - f_min) + f_min
        return df

    def unscaleMinMax(self, data=None, columns=None, columns_mapping={}):
        """Unscale from min-max.
        Retreives data from the same range as the original features.

        :param data: Data set to unscale, defaults to None, takes data provided in :py:meth:`__init__`.
        :type data: :obj:`pandas.DataFrame`, optional
        :param columns: Columns to be unscaled, defaults to None, takes the list provided in :py:meth:`__init__`.
        :type columns: :obj:`list`, optional
        :param columns_mapping: Mapping between eventual renamed columns and original scaled column.
        :type columns_mapping: :obj:`dict`, optional

        :formula: .. math:: x_{\\text{unscaled}} = x_{\\text{scaled}} \\cdot \\left(\\max(x) - \\min(x) \\right) + \\min(x)

        :return: Unscaled data set.
        :rtype: :obj:`pandas.DataFrame`
        """
        cols = self._col_to_list(columns)
        df = self.data if data is None else data.copy()
        unscale_suffix = '_unscaled' if self._minmax_suffix != '' else ''
        f_min, f_max = self._minmax_feature_range

        for c in cols:
            # Apply eventual column name mapping
            _c = c if columns_mapping.get(c) is None else columns_mapping.get(c)
            # Retreive min and max
            _min = self.scalers[_c]['min']
            _max = self.scalers[_c]['max']
            tmp = (df[c + self._minmax_suffix] - f_min) / (f_max - f_min)
            df[c + unscale_suffix] = (tmp * (_max - _min)) + _min
        return df

    def Normalize(self, center=True, reduce=True, data=None, columns=None, col_suffix=''):
        """Data normalizer.
        Centers and reduces features (from mean and standard deviation).

        :param center: Center the variable, i.e. substract the mean, defaults to True.
        :type center: :obj:`bool`, optional
        :param reduce: Reduce the variable, i.e. devide by standard deviation, defaults to True.
        :type reduce: :obj:`bool`, optional
        :param data: Data set to normalize, defaults to None, takes data provided in :py:meth:`__init__`.
        :type data: :obj:`pandas.DataFrame`, optional
        :param columns: Columns to be normalize, defaults to None, takes the list provided in :py:meth:`__init__`.
        :type columns: :obj:`list`, optional
        :param col_suffix: [description], defaults to ''
        :type col_suffix: :obj:`str`, optional

        :formula: .. math:: x_{\\text{scaled}} = \\dfrac{x - \\bar{x}}{\\sqrt{\\mathbb{V}(x)}}

        :return: Data set with normalized features.
        :rtype: :obj:`pandas.DataFrame`
        """
        self._standard_suffix = col_suffix
        cols = self._col_to_list(columns)
        df = self.data if data is None else data.copy()

        for c in cols:
            # Retreive mean
            if center:
                _mean = self.scalers[c]['mean']
                self.scalers[c].update({'center': True})
            else:
                _mean = 0.
                self.scalers[c].update({'center': False})
            # Retreive std
            if reduce:
                _std = self.scalers[c]['std']
                self.scalers[c].update({'reduce': True})
            else:
                _std = 1.
                self.scalers[c].update({'reduce': False})

            df[c + col_suffix] = (df[c] - _mean) / _std
        return df

    def unscaleNormalize(self, data=None, columns=None, columns_mapping={}):
        """Denormalize data to retreive the same range as the original data set.

        :param data: Data set to unscale, defaults to None, takes data provided in :py:meth:`__init__`.
        :type data: :obj:`pandas.DataFrame`, optional
        :param columns: Columns to be unscaled, defaults to None, takes the list provided in :py:meth:`__init__`.
        :type columns: :obj:`list`, optional
        :param columns_mapping: Mapping between eventual renamed columns and original scaled column.
        :type columns_mapping: :obj:`dict`, optional

        :formula: .. math:: x_{\\text{unscaled}} = x_{\\text{scaled}} \\cdot \\sqrt{\\mathbb{V}(x)} + \\bar{x}

        :return: De-normalized data set.
        :rtype: :obj:`pandas.DataFrame`
        """
        cols = self._col_to_list(columns)
        df = self.data if data is None else data.copy()
        unscale_suffix = '_unscaled' if self._standard_suffix != '' else ''

        for c in cols:
            # Apply eventual column name mapping
            _c = c if columns_mapping.get(c) is None else columns_mapping.get(c)
            # Retreive min and max
            _mean = self.scalers[_c]['mean'] if self.scalers[_c]['center'] else 0.
            _std = self.scalers[_c]['std'] if self.scalers[_c]['reduce'] else 1.

            df[c + unscale_suffix] = (df[c + self._standard_suffix] * _std) + _mean

        return df



import warnings
import math
import numpy as np
import pandas as pd
from scipy.stats import norm

# Create custom warning: Value
class ValueWarning(UserWarning):
    pass

# Create custom warning: Convergence
class ConvergenceWarning(UserWarning):
    pass


def get_significance(proba):
    if proba < 0.001:
        return('***')
    elif proba < 0.01:
        return(' **')
    elif proba < 0.05:
        return('  *')
    elif proba < 0.1:
        return('  .')
    else:
        return('   ')


def summary(s):
    """Print summary data frame to string

    :param s: Summary from model as dataframe.
    :type s: :obj:`pandas.DataFrame`

    :return: Formatted summary
    :rtype: :obj:`str`
    """
    max_var = np.max([len(v) for v in s.Variables])

    add_sp = ' ' * np.max([max_var - 17, 0])
    add_sep = '=' * np.max([max_var - 17, 0])
    space = np.max([max_var, 17])

    summ = f"============================================================================================================={add_sep}\n"
    summ += f"| Variables        {add_sp} | Coefficients   | Std. Errors  | t-values   | 95% Conf Int.            | Probabilities |\n"
    summ += f"============================================================================================================={add_sep}\n"

    for i in range(s.shape[0]):
        vari = s.Variables[i]
        c = s.Coefficients[i]
        se = s['Standard Errors'][i]
        tv = s['t-values'][i]
        pb = s['Probabilities'][i]
        sign = s['Significance'][i]
        ci_lo = s['CI_lo'][i]
        ci_hi = s['CI_hi'][i]

        c_val = round(c, 5)
        std_val = round(se, 5)
        tv_val = round(tv, 3)
        pb_val = round(pb, 3) if math.fabs(pb) < 1000 else round(pb, 3 - len(str(int(pb))))
        ci_lo = f'{round(ci_lo, 3):9}'
        ci_hi = f'{round(ci_hi, 3):9}'

        summ += f"| {vari:{space}} |  {c_val:13} | {std_val:12} | {tv_val:10} | [{ci_lo:10}, {ci_lo:10}] |  {pb_val:6}   {sign:3} |\n"
    summ += f"============================================================================================================={add_sep}\n"
    summ += f"| Significance codes: 0. < *** < 0.001 < ** < 0.01 < * < 0.05 < . < 0.1 < '' < 1                            {add_sp}|\n"
    summ += f"============================================================================================================={add_sep}\n"
    return summ


def test_summary(df, critical_value, t_value, p_value, alpha=0.05, title='', h0='H0', h1='H0 does not hold', extra='', h0_conclu='', h1_conclu=''):

    # Format for output

    blanks = int((58 - len(title)) / 2)
    right_space = ' ' * np.max([int(blanks), 0])
    left_space = ' ' * np.max([58 - len(title) - blanks, 0])
    dfree = round(df, 10 - len(str(int(df))))
    cri_value = round(critical_value, 11 - len(str(int(critical_value))))
    t_val = round(t_value, 8 - len(str(int(t_value))))
    p_v = round(p_value, 6 - len(str(int(p_value))))
    r = 'True' if alpha < p_value else 'False'

    if alpha < p_value:
        conclusion = ' * We cannot reject H0: ' + h0 + '\n'
    else:
        conclusion = ' * We reject H0, hence ' + h1 + '\n'

    summ = "+------------------------------------------------------------+\n"
    summ += f"|{right_space} {title} {left_space}|\n"
    summ += "+------------+----------------+------------+---------+-------+\n"
    summ += "|     df     | Critical value | Stat value | p-value |   H0  |\n"
    summ += "+------------+----------------+------------+---------+-------+\n"
    summ += f"| {dfree:10} | {cri_value:14} | {t_val:10} | {p_v:7} | {r:5} |\n"
    summ += "+------------+----------------+------------+---------+-------+\n"
    summ += conclusion
    summ += h0_conclu if r == 'True' else ''
    summ += h1_conclu if r == 'False' else ''
    summ += extra

    return summ


def _to_array(x, name='x'):

    warnings.filterwarnings('ignore')

    if type(x) in [pd.Series, pd.DataFrame]:
        return np.array(x.values)
    elif type(x) in [list]:
        return np.array(x)
    elif type(x) == np.ndarray:
        return x
    else:
        raise TypeError(f'Type for {name} is not valid.')
    warnings.filterwarnings('default')


def _to_list(x, name='x'):
    warnings.filterwarnings('ignore')
    if type(x) in [pd.Series, pd.DataFrame]:
        return list(x.values)
    elif type(x) == list:
        return x
    elif type(x) == np.ndarray:
        if x.shape == (len(x), 1):
            return [x[0] for x in np.asarray(x)]
        elif x.shape == (len(x),):
            return list(x)
        else:
            raise TypeError(f'Cannot properly read shape for {name}.')
    else:
        raise TypeError(f'Type for {name} is not valid.')
    warnings.filterwarnings('default')


def format_object(x, to_type='array', name='x'):
    if to_type in ['array', 'ndarray', 'to_array', 'to array']:
        return _to_array(x, name=name)
    elif to_type in ['list', 'to_list', 'to list']:
        return _to_list(x, name=name)
    else:
        raise ValueError(f'Value for to_type is not value. Can either be array or list. Got {to_type}.')



import numpy as np
from scipy import stats as scp
import pandas as pd
import math
from scipy.stats import norm
import matplotlib.pyplot as plt

from ..data.ProcessData import parse_formula
from ..misc import summary, get_significance
from ..nonparametrics.kernels import gaussian

# TODO: Add Log-Likehood + AIC + BIC to OLS
# TODO: Add dask for GPU usage
# TODO: Add LinearBayes with sigma unknown @ybendou


class OLS:

    def __init__(self, formula, data, fit_intercept=False):
        """Ordinary Least Squares regression.

        :param formula: Regression formula to be run of the form :obj:`y ~ x1 + x2`. See :func:`~parse_formula` in
                        `ProcessData <../../data/process.html#statinf.data.ProcessData.parse_formula>`_
        :type formula: :obj:`str`
        :param data: Input data with Pandas format.
        :type data: :obj:`pandas.DataFrame`
        :param fit_intercept: Used for adding intercept in the regression formula, defaults to False.
        :type fit_intercept: :obj:`bool`, optional
        """

        super(OLS, self).__init__()
        # Parse formula
        df, self.X_col, self.Y_col = parse_formula(data=data, formula=formula, check_values=True, return_all=True)
        # self.no_space_formula = formula.replace(' ', '')
        # self.Y_col = self.no_space_formula.split('~')[0]
        # self.X_col = self.no_space_formula.split('~')[1].split('+')
        self.formula = formula
        # Subset X
        self.X = df[self.X_col].to_numpy()
        # Target variable
        self.Y = df[self.Y_col].to_numpy()
        # Degrees of freedom of the population
        self.dft = self.X.shape[0]
        # Degree of freedom of the residuals
        self.dfe = self.X.shape[0] - self.X.shape[1]
        # Size of the population
        self.n = self.X.shape[0]
        # Number of explanatory variables / estimates
        self.p = self.X.shape[1]
        # Use intercept or only explanatory variables
        self.fit_intercept = fit_intercept
        # Beta standard errors for confidence intervals
        self.std_err = None

    def _get_X(self):
        if self.fit_intercept:
            return(np.hstack((np.ones((self.n, 1), dtype=self.X.dtype), self.X)))
        else:
            return(self.X)

    def get_betas(self):
        """Computes the estimates for each explanatory variable

        :formula: .. math:: \\beta = (X'X)^{-1} X'Y

        :return: Estimated coefficients
        :rtype: :obj:`numpy.array`
        """
        XtX = self._get_X().T.dot(self._get_X())
        XtX_1 = np.linalg.inv(XtX)
        XtY = self._get_X().T.dot(self.Y)
        beta = XtX_1.dot(XtY)
        return(beta)

    def fitted_values(self):
        """Computes the estimated values of Y

        :formula: .. math:: \\hat{Y} = X \\beta

        :return: Fitted values for Y.
        :rtype: :obj:`numpy.array`
        """
        betas = self.get_betas()
        Y_hat = np.zeros(len(self.Y))
        for i in range(len(betas)):
            Y_hat += (betas[i] * self._get_X().T[i])
        return(Y_hat)

    def _get_error(self):
        """Compute the error term/residuals

        :formula: .. math:: \\epsilon = Y - \\hat{Y}

        :return: Estimated residual term.
        :rtype: :obj:`numpy.array`
        """
        res = self.Y - self.fitted_values()
        return(res)

    def rss(self):
        """Residual Sum of Squares

        :formula: .. math:: RSS = \\sum_{i=1}^{n} (y_{i} - \\hat{y}_{i})^{2}

            where :math:`y_{i}` denotes the true/observed value of :math:`y` for individual :math:`i`
            and :math:`\\hat{y}_{i}` denotes the predicted value of :math:`y` for individual :math:`i`.

        :return: Residual Sum of Squares.
        :rtype: :obj:`float`
        """
        return((self._get_error()**2).sum())

    def tss(self):
        """Total Sum of Squares

        :formula: .. math:: TSS = \\sum_{i=1}^{n} (y_{i} - \\bar{y})^{2}

            where :math:`y_{i}` denotes the true/observed value of :math:`y` for individual :math:`i`
            and :math:`\\bar{y}_{i}` denotes the average value of :math:`y`.

        :return: Total Sum of Squares.
        :rtype: :obj:`float`
        """

        y_bar = self.Y.mean()
        total_squared = (self.Y - y_bar) ** 2
        return(total_squared.sum())

    def r_squared(self):
        """:math:`R^{2}` -- Goodness of fit

        :formula: .. math:: R^{2} = 1 - \\dfrac{RSS}{TSS}

        :return: Goodness of fit.
        :rtype: :obj:`float`
        """

        return(1 - self.rss() / self.tss())

    def adjusted_r_squared(self):
        """Adjusted-:math:`R^{2}` -- Goodness of fit

        :formula: .. math:: R^{2}_{adj} = 1 - (1 - R^{2}) \\dfrac{n - 1}{n - p - 1}

            where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size

        :references: Theil, Henri (1961). Economic Forecasts and Policy.

        :return: Adjusted goodness of fit.
        :rtype: :obj:`float`
        """

        adj_r_2 = 1 - (1 - self.r_squared()) * (self.n - 1) / (self.n - self.p - 1)
        return(adj_r_2)

    def _fisher(self):
        """Fisher test

        :formula: .. math:: \\mathcal{F} = \\dfrac{TSS - RSS}{\\frac{RSS}{n - p}}

            where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size

        :references: Shen, Q., & Faraway, J. (2004). `An F test for linear models with functional responses
                     <https://www.jstor.org/stable/24307230>`_. Statistica Sinica, 1239-1257.

        :return: Value of the :math:`\\mathcal{F}`-statistic.
        :rtype: :obj:`float`
        """
        MSE = (self.tss() - self.rss()) / (self.p - 1)
        MSR = self.rss() / self.dfe
        return(MSE / MSR)

    def _std_err(self):
        """Standard error function

        :formula:

            .. math:: \\mathbb{V}(\\beta) = \\sigma^{2} X'X

            where :math:`\\sigma^{2} = \\frac{RSS}{n - p -1}`

        :return: Standard error of the estimates.
        :rtype: :obj:`numpy.array`
        """
        X = self._get_X()

        sigma_2 = (sum((self._get_error())**2)) / (len(X) - len(X[0]))
        variance_beta = sigma_2 * (np.linalg.inv(np.dot(X.T, X)).diagonal())
        self.std_err = np.sqrt(variance_beta)
        return self.std_err

    def _loglikelihood(self):
        """Standard error function

        :formula:

            .. math:: l = \\dfrac{n}{2} \\log{2 \\pi} - \\dfrac{n}{2} \\log{\\dfrac{RSS}{n}} - \\dfrac{n}{2}

        :return: Log-likelihood.
        :rtype: :obj:`float`
        """
        _ll = - (self.n / 2) * np.log(2 * math.pi) - (self.n / 2) * np.log(self.rss() / self.n) - (self.n / 2)
        self.loglikelihood = _ll
        return _ll

    def _aic(self, metric='aic'):
        """Akaike Information Criterion and Bayesian Information Criterion

        :param metric: Define what metric to return, defaults to 'aic'.
        :type return_df: :obj:`str`

        :formula: * AIC:

            .. math:: AIC = - 2 \\log{L(\\theta)} + 2 p

            * BIC:

            .. math:: AIC = - 2 \\log{L(\\theta)} + \\log{n} p

            where :math:`p` is the number of covariates and :math:`L` the likelihood function.

        :references: * Cameron, A. C., & Trivedi, P. K. (2009). Microeconometrics using stata (Vol. 5, p. 706). College Station, TX: Stata press.

        :return: Information criterion.
        :rtype: :obj:`float`
        """
        self.aic = -2 * self._loglikelihood() + 2 * self.p
        self.bic = -2 * self._loglikelihood() + np.log(self.n) * self.p

        if metric.lower() == 'bic':
            return self.bic
        else:
            return self.aic

    def summary(self, return_df=False):
        """Statistical summary for OLS

        :param return_df: Return the summary as a Pandas DataFrame, else returns a string, defaults to False.
        :type return_df: :obj:`bool`

        :formulae: * Fisher test:

            .. math:: \\mathcal{F} = \\dfrac{TSS - RSS}{\\frac{RSS}{n - p}}

            where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size


            * Covariance matrix:

            .. math:: \\mathbb{V}(\\beta) = \\sigma^{2} X'X

            where :math:`\\sigma^{2} = \\frac{RSS}{n - p -1}`


            * Coefficients' significance:

            .. math:: p = 2 \\left( 1 - T_{n} \\left( \\dfrac{\\beta}{\\sqrt{\\mathbb{V}(\\beta)}} \\right) \\right)

            where :math:`T` denotes the Student cumulative distribution function (c.d.f.) with :math:`n` degrees of freedom


        :references: * Student. (1908). The probable error of a mean. Biometrika, 1-25.
            * Shen, Q., & Faraway, J. (2004). `An F test for linear models with functional responses <https://www.jstor.org/stable/24307230>`_.
              Statistica Sinica, 1239-1257.
            * Wooldridge, J. M. (2016). `Introductory econometrics: A modern approach
              <https://faculty.arts.ubc.ca/nfortin/econ495/IntroductoryEconometrics_AModernApproach_FourthEdition_Jeffrey_Wooldridge.pdf>`_.
              Nelson Education.
            * Cameron, A. C., & Trivedi, P. K. (2009). Microeconometrics using stata (Vol. 5, p. 706). College Station, TX: Stata press.

        :return: Model's summary.
        :rtype: :obj:`pandas.DataFrame` or :obj:`str`
        """
        # Initialize
        betas = self.get_betas()
        X = self._get_X()

        self.std_err = self._std_err()
        t_values = betas / self.std_err

        p_values = [2 * (1 - scp.t.cdf(np.abs(i), (len(X) - 1))) for i in t_values]
        z = norm.ppf(0.975)

        summary_df = pd.DataFrame()
        summary_df["Variables"] = ['(Intercept)'] + self.X_col if self.fit_intercept else self.X_col
        summary_df["Coefficients"] = betas
        summary_df["Standard Errors"] = self.std_err
        summary_df["t-values"] = t_values
        summary_df["Probabilities"] = p_values
        summary_df["Significance"] = summary_df["Probabilities"].map(lambda x: get_significance(x))
        summary_df["CI_lo"] = summary_df["Coefficients"] - z * summary_df["Standard Errors"]
        summary_df["CI_hi"] = summary_df["Coefficients"] + z * summary_df["Standard Errors"]

        _r2 = round(self.r_squared(), 5)
        ar2 = round(self.adjusted_r_squared(), 5)
        _n_ = self.n
        _p_ = self.p
        #
        fis = round(self._fisher(), 3)
        llf = round(self._loglikelihood(), 3)
        aic = round(self._aic(), 3)

        #
        if return_df:
            return(summary_df)
        else:
            max_var = np.max([len(v) for v in summary_df.Variables])

            add_sp = ' ' * np.max([max_var - 17, 0])
            add_sep = '=' * np.max([max_var - 17, 0])

            summ = f"============================================================================================================={add_sep}\n"
            summ += f'|                                                OLS summary                                                {add_sp}|\n'
            summ += f"============================================================================================================={add_sep}\n"
            summ += f"| n                   =                  {_n_:10} | p                 =                        {_p_:10} {add_sp}|\n"
            summ += f"| R²                  =                  {_r2:10} | R² Adj.           =                        {ar2:10} {add_sp}|\n"
            summ += f"| Log-likelihood      =             {llf:15} | AIC               =                        {aic:10} {add_sp}|\n"
            summ += f"| Fisher value        =             {fis:15} |                                                       {add_sp}|\n"
            summ += summary(summary_df)
            return(summ)

    def predict(self, new_data, conf_level=None):
        """Predicted :math:`\\hat{Y}` values for for a new dataset

        :param new_data: New data to evaluate with pandas data-frame format.
        :type new_data: :obj:`pandas.DataFrame`
        :param conf_level: Level of the confidence interval, defaults to None.
        :type conf_level: :obj:`float`

        :formulae: .. math:: \\hat{Y} = X \\hat{\\beta}

        The confidence interval is computed as:

            .. math:: \\left[ \\hat{Y} \\pm z_{1 - \\frac{\\alpha}{2}} \\dfrac{\\sigma}{\\sqrt{n - 1}} \\right]

        :return: Predictions :math:`\\hat{Y}`
        :rtype: :obj:`numpy.array`
        """
        df = parse_formula(data=new_data, formula=self.formula, check_values=True)
        X_array = df[self.X_col].to_numpy()
        if self.fit_intercept:
            new_X = np.hstack((np.ones((df.shape[0], 1), dtype=X_array.dtype), X_array))
        else:
            new_X = X_array

        y_pred = np.dot(new_X, self.get_betas())

        # No CI required
        if conf_level is None:
            pred = y_pred
        else:
            assert (conf_level > 0.) & (conf_level < 1.), "Your confidence level needs to be between 0 and 1."
            # User needs CI
            if self.std_err is None:
                _ = self.summary(return_df=True)

            quant_order = 1 - ((1 - conf_level) / 2)
            cv = scp.norm.ppf(quant_order)
            pred = pd.DataFrame({'Prediction': y_pred,
                                 'LowerBound': y_pred - cv * (self.std_err.sum() / np.sqrt(self.n - 1)),
                                 'UpperBound': y_pred + cv * (self.std_err.sum() / np.sqrt(self.n - 1))})

        return pred


class LinearBayes:

    def __init__(self):
        """
            Class for a bayesian linear regression with **known standard deviation** of the residual distribution.

            .. warning:: This function is still under development.
                This is a beta version, please be aware that some functionalitie might not be available.
                The full stable version soon will be released.

            :param w_0: mean of the prior distribution of the weights assuming it's a gaussian, default is 0.
            :type  w_0: :obj:`numpy.array`
            :param V_0: covariance matrix of the prior distribution of the weights, default is the identity matrix.
            :type  V_0: :obj:`numpy.array`
            :param w_n: mean of the posterior distribution of the weights after observing the data.
            :type  w_n: :obj:`numpy.array`
            :param V_n: covariance matrix the posterior distribution of the weights after observing the data.
            :type  V_n: :obj:`numpy.array`

            :references: * Murphy, K. P. (2012). Machine learning: a probabilistic perspective.

            :source: Inspired by: https://jessicastringham.net/2018/01/03/bayesian-linreg/
        """
        self.w_0 = None
        self.V_0 = None
        self.w_n = None
        self.V_n = None

    def fit(self, X, y, true_sigma=None, w_0=None, V_0=None):
        """
            Fits the data using a linear regression model, finds the posterior distribution of the weights given the std of the residuals known.
            The case with std unknown will be added in the next implementation.

            :param X: Input data.
            :type  X: :obj:`numpy.array`
            :param y: Data values.
            :type  y: :obj:`numpy.array`
            :param true_sigma: Standard error of the residual distribution.
            :type  true_sigma: :obj:`float`
            :param w_0: Mean of the prior distribution of the weights assuming it's a gaussian, default is 0.
            :type  w_0: :obj:`float`
            :param V_0: Covariance matrix of the prior distribution of the weights, default is the identity matrix.
            :type  V_0: :obj:`float`
        """
        # Setting a gaussian prior of N(0,1) in case no prior is refered
        if type(w_0) != np.ndarray and type(w_0) != list:
            w_0 = np.hstack([0] * (X.shape[1] + 1))
        if type(V_0) != np.ndarray and type(V_0) != list:
            V_0 = np.diag([1] * (X.shape[1] + 1))**2

        w_0 = w_0[:, None]

        # Creating a new column for the bias term
        phi = np.hstack((
            np.ones((X.shape[0], 1)), X
        ))

        assert true_sigma is not None, "Error, please specify a value for true_sigma"

        sigma_y = true_sigma  # True std of the generated data

        inv_V_0 = np.linalg.inv(V_0)

        V_n = sigma_y**2 * np.linalg.inv(sigma_y**2 * inv_V_0 + (phi.T @ phi))
        w_n = V_n @ inv_V_0 @ w_0 + 1 / (sigma_y**2) * V_n @ phi.T @ y

        self.w_0 = w_0
        self.V_0 = V_0
        self.w_n = w_n
        self.V_n = V_n

    def plot_weight_distributions(self, res=100, xlim=(-8, 8), ylim=(-8, 8)):
        """
            Plots the weight distribution for the prior and the posterior probabilities.

            :param res: Resolution of the grid, default is 100.
            :type res: :obj:`int`
            :param xlim: Tuple of the min and max values of the grid along the x axis, default is (-8,8).
            :type xlim: :obj:`tuple`
            :param ylim: Tuple of the min and max values of the grid along the y axis, default is (-8,8).
            :type ylim: :obj:`tuple`
        """

        x = np.linspace(xlim[0], xlim[1], res)
        y = np.linspace(ylim[0], ylim[1], res)
        xx, yy = np.meshgrid(x, y)
        xxyy = np.c_[xx.ravel(), yy.ravel()]

        zz_0 = gaussian(xxyy, self.w_0.reshape(1, -1), self.V_0).reshape(1, -1)[0].round(15)
        zz_n = gaussian(xxyy, self.w_n.reshape(1, -1), self.V_n).reshape(1, -1)[0].round(15)
        zz_0[zz_0 == 0] = np.nan
        zz_n[zz_n == 0] = np.nan

        # reshape and plot image

        img_0 = zz_0.reshape(res, res)
        img_n = zz_n.reshape(res, res)
        plt.contourf(xx, yy, img_0, alpha=0.5, cmap=plt.cm.RdBu)
        plt.contourf(xx, yy, img_n, alpha=0.5, cmap=plt.cm.RdBu)
        plt.title('Posterior and prior distributions of the weights')

    def plot_posterior_line(self, X, y, n_lines=200, res=100, xlim=(-1, 10)):
        """
            Plots the model's distribution sampled from the posterior distribution of the weights.

            :param X: Input data.
            :type  X: :obj:`numpy.array`
            :param y: Data values.
            :type  y: :obj:`numpy.array`
            :param n_lines: Number of lines sampled from the posterior distribution, default is 200.
            :type n_lines: :obj:`int`
            :param res: Resolution of the grid, default is 100.
            :type res: :obj:`int`
            :param xlim: Tuple of the min and max values of the grid along the x axis, default is (-1,10).
            :type xlim: :obj:`tuple`
        """

        w_posterior = np.random.randn(n_lines, self.w_n.shape[0]) @ self.V_n + self.w_n.reshape(1, -1)
        X_grid = np.linspace(xlim[0], xlim[1], res)
        X_grid = np.vstack((
            np.ones(X_grid.shape[0]),
            X_grid
        )).T

        y_posterior = X_grid @ w_posterior.T
        plt.fill_between(X_grid[:, 1], y_posterior.min(axis=1), y_posterior.max(axis=1), alpha=0.5)
        plt.scatter(X, y)
        plt.title(f'{n_lines} lines sampled from the posterior distribution after including the data')
        plt.show()



import numpy as np
import theano.tensor as T

# Default activation functions
def sigmoid(x):
    """Sigmoid activation function

    :param x: Input value
    :type x: float or numpy.array

    :return: Sigmoid activated value: :math:`sigmoid(x) = \\dfrac{1}{1 + e^{-x}}`
    :rtype: float
    """
    return 1 / (1 + np.exp(-x))


def tanh(x):
    """Hyperbolic tangent activation function

    :param x: Input value
    :type x: float or numpy.array

    :return: Activated value: :math:`tanh(x)`
    :rtype: float
    """
    return np.tanh(x)


def relu(x):
    """Rectified Linear Units activation function

    :param x: Input value
    :type x: float or numpy.array

    :return: Activated value: :math:`relu(x) = \\max(0, x)`
    :rtype: float
    """
    return np.maximum(0, x)


def softplus(x):
    """Softplus activation function

    :param x: Input value
    :type x: float or numpy.array

    :return: Activated value: :math:`softplus(x) = \\log(1 + e^{-x})`
    :rtype: float
    """
    return np.log(1 + np.exp(-x))


def logit(x, weights, bias=0, tensor=False):
    """Logistic function

    :param x: Input value
    :type x: numpy.array
    :param weights: Vector of weights :math:`\\beta`
    :type weights: numpy.array
    :param bias: Vector of bias :math:`\\epsilon`, defaults to 0.
    :type bias: numpy.array
    :param tensor: Perform computation as tensor (theano type), defaults to False
    :type tensor: bool, optional

    :return: Logistic transformation: :math:`logit(x, \\beta) = \\dfrac{1}{1 + e^{-x \\beta}}`
    :rtype: float
    """

    if tensor:
        return 1 / (1 + T.exp(-T.dot(x, weights) + bias))
    else:
        return 1 / (1 + np.exp(-x.dot(weights) + bias))



import warnings
import pandas as pd
import numpy as np
import math

from ..misc import format_object


class BinaryPerformance:
    def __init__(self, y_true, y_pred):
        """Gives detailed perfomance metrics for binary calssification models.

        :param y_true: Array of true targets.
        :type y_true: :obj:`numpy.ndarray`
        :param y_pred: Array of predicted targets.
        :type y_pred: :obj:`numpy.ndarray`
        """
        warnings.filterwarnings('ignore')
        # Format y_true and y_pred
        true = format_object(y_true, to_type='list', name='y_true')
        pred = format_object(y_pred, to_type='list', name='y_pred')

        # Put data to a DF
        for_conf = pd.DataFrame({'true': true,
            'pred': pred,
            'perf': ''})

        # Compute True and False positives/negatives
        for_conf['perf'][(for_conf.true == 1) & (for_conf.pred == 1)] = 'true_positive'
        for_conf['perf'][(for_conf.true == 0) & (for_conf.pred == 0)] = 'true_negative'
        for_conf['perf'][(for_conf.true == 1) & (for_conf.pred == 0)] = 'false_negative'
        for_conf['perf'][(for_conf.true == 0) & (for_conf.pred == 1)] = 'false_positive'
        # Number of True and False positives/negatives
        self.tp = len(for_conf[(for_conf.perf == 'true_positive')])
        self.tn = len(for_conf[(for_conf.perf == 'true_negative')])
        self.fn = len(for_conf[(for_conf.perf == 'false_negative')])
        self.fp = len(for_conf[(for_conf.perf == 'false_positive')])
        self.m = len(y_pred)
        # Confusion matrix
        self.conf = pd.DataFrame({'True 0': [0., 0.], 'True 1': [0., 0.]}, index=['Predicted 0', 'Predicted 1'])
        # Fill confusion matrix
        self.conf['True 0']['Predicted 0'] = self.tn/self.m
        self.conf['True 0']['Predicted 1'] = self.fp/self.m
        self.conf['True 1']['Predicted 0'] = self.fn/self.m
        self.conf['True 1']['Predicted 1'] = self.tp/self.m
        warnings.filterwarnings('default')

    def accuracy(self):
        """Binary accuracy of the model. Percentage of equal values between :obj:`y_true` and :obj:`y_pred`.

        :formula: .. math:: accuracy = \\dfrac{TP + TN}{n}

        :return: Accuracy
        :rtype: :obj:`float`
        """
        return((self.tp + self.tn)/self.m)

    def confusion(self):
        """Confusion matrix

        :return: Confusion matrix

            +-----------------+------------+------------+
            |                 | **True 0** | **True 0** |
            +=================+============+============+
            | **Predicted 0** | :math:`TN` | :math:`TN` |
            +-----------------+------------+------------+
            | **Predicted 1** | :math:`FP` | :math:`TP` |
            +-----------------+------------+------------+

        :rtype: :obj:`pandas.DataFrame`
        """
        return(self.conf * 100)

    def precision(self):
        """Precision metric, proportion of actual 1 values amongst the ones predicted.

        :formula: .. math:: precision = \\dfrac{TP}{TP + FP}

        :return: Precision
        :rtype: :obj:`float`
        """
        return(self.tp / (self.tp + self.fp))

    def recall(self):
        """  Recall metric, proportion of values we predicted as one from the actuals ones.

        :formula: .. math:: recall = \\dfrac{TP}{TP + FN}

        :return: Recall
        :rtype: :obj:`float`
        """
        return(self.tp / (self.tp + self.fn))

    def F1_score(self):
        """F1-score

        :formula: .. math:: F_{1} = 2 \\cdot \\dfrac{precision \\times recall}{precision + recall}

        :return: F1-score
        :rtype: :obj:`float`
        """
        return(2 * (self.precision() * self.recall()) / (self.precision() + self.recall()))


def mean_squared_error(y_true, y_pred, root=False):
    """Mean Squared Error

    :param y_true: Real values on which to compare.
    :type y_true: :obj:`numpy.ndarray`
    :param y_pred: Predicted values.
    :type y_pred: :obj:`numpy.ndarray`
    :param root: Return Root Mean Squared Error (RMSE), defaults to False.
    :type root: :obj:`bool`, optional

    :formula: :math:`loss = \\dfrac{1}{m} \\times \\sum_{i=1}^{m} (y_i - \\hat{y}_i)^2`

    :references: * Friedman, J., Hastie, T. and Tibshirani, R., 2001. `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_. Ch. 2, pp. 24.

    :return: Mean Squared Error or its root.
    :rtype: float
    """

    true = format_object(y_true, to_type='array', name='y_true')
    pred = format_object(y_pred, to_type='array', name='y_pred')
    # pred = to_array(y_pred, 'y_pred')

    # Compute the square of the difference
    loss = (pred - true)**2

    if root:
        return math.sqrt(loss.mean())
    else:
        return loss.mean()


# MAPE formula
def mape(y_true, y_pred, weights=False):
    """Computes the Mean Absolute Percentage Error (MAPE) or Weighted Mean Absolute Percentage Error (WMAPE).

    :param y_true: Real values on which to compare.
    :type y_true: :obj:`numpy.array`
    :param y_pred: Predicted values.
    :type y_pred: :obj:`numpy.array`
    :param weights: Compute WMAPE.
    :type weights: :obj:`bool`

    :formula:

    * :math:`MAPE(y, \\hat{y}) = \\dfrac{100}{n} \\sum_{i=1}^{n} \\dfrac{|y - \\hat{y}|}{y}`
    * :math:`WMAPE(y, \\hat{y}) = 100 \\dfrac{\\sum_{i=1}^{n} {\\dfrac{|y - \\hat{y}|}{y}} \\times y}{\\sum_{i=1}^{n} y}`

    :return: Mean Absolute Percentage Error as percentage.
    :rtype: float
    """

    warnings.filterwarnings('ignore')
    # Format y_true and y_pred
    _true = format_object(y_true, to_type='list', name='y_true')
    _pred = format_object(y_pred, to_type='list', name='y_pred')

    preds = pd.DataFrame({'true': _true, 'pred': _pred})
    preds['abs_err'] = np.abs(preds['true'] - preds['pred'])
    preds['err_contrib'] = preds['abs_err']/preds['true']

    if weights:
        preds['err_contrib'] = preds['err_contrib'] * preds['true'] * 100
        out = preds['err_contrib'].sum() / preds.true.sum()
    else:
        out = 100 * preds['err_contrib'].sum() / preds['true'].sum()

    return out



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import random

from ..nonparametrics.kernels import gaussian


class KMeans():
    def __init__(self, k=1, max_iter=100, init='random', random_state=0):
        """
            K-means clustering implementation.

            .. warning:: This function is still under development.
                This is a beta version, please be aware that some functionalitie might not be available.
                The full stable version soon will be released.

            :param k: number of clusters, default is 1.
            :type k: :obj:`int`
            :param max_iter: number of iterations for convergence.
            :type max_iter: :obj:`int`
            :param init: initialization option, options are random or kmeans++ .
            :type init: :obj:`String`
            :param random_state: seed of the random state, default is 0.
            :type random_state: :obj:`int`
            :param labels: labels for each datapoint.
            :type labels: :obj:`numpy.array`
            :param centroids: coordinates of the centroids.
            :type centroids: :obj:`numpy.array`

            :references:
                * Friedman, J., Hastie, T., & Tibshirani, R. (2001). The elements of statistical learning (Vol. 1, No. 10).
                  New York: Springer series in statistics.
        """

        self.max_iter = max_iter
        self.init = init
        self.random_state = random_state
        self.k = k
        self.labels_ = 0
        self.centroids = 0

    def fit(self, X):
        """
            Fit the model to the data using different initializations (random init or kmeans++)

            :param X: Input data.
            :type X: :obj:`numpy.array`
        """
        # Initialize centroids randomly or using k-means ++
        if type(X) == pd.core.frame.DataFrame:
            data = X.values
        elif type(X) == np.ndarray:
            data = X
        else:
            raise TypeError('X should be either a numpy array or a pandas DataFrame')

        if self.init == 'random':
            centroids = data.copy()
            r = np.random.RandomState(seed=self.random_state)
            r.shuffle(centroids)
            centroids = centroids[:self.k, :]
        elif self.init == 'kmeans++':
            candidates = data.copy()[:, :2]
            # Pick a first random centroid
            random_candidate_index = np.random.randint(0, candidates.shape[0] - 1)
            # Add it to the list of centroids
            centroids = [candidates[random_candidate_index, :]]
            # Remove the picked centroid from the list of potential centroids
            candidates = np.delete(candidates, random_candidate_index, axis=0)
            for k in range(self.k - 1):
                distances = self.get_distance(candidates, np.array(centroids).mean(axis=0).reshape(1, -1))[0]
                # Normalize distances to get the probabilities
                probabilities = distances / distances.sum()
                # Pick new centroid with a probability proportional to the distance
                random_candidate_index = np.random.choice(np.arange(0, candidates.shape[0]), p=probabilities)
                centroids.append(candidates[random_candidate_index, :])
                # Remove the picked centroid from the list of potential centroids
                candidates = np.delete(candidates, random_candidate_index, axis=0)
            centroids = np.array(centroids)

        for _l in range(0, self.max_iter):
            labels_ = self.closest_centroid(data, centroids)  # Find index of closest centroid to each point
            centroids = self.move_centroids(data, labels_, centroids)  # Update the values of the new centroid

        self.labels_ = self.closest_centroid(data, centroids)
        self.centroids = centroids

    def closest_centroid(self, points, centroids):
        """
            Returns an array containing the index to the nearest centroid for each point

            :param points: features of each point.
            :type points: :obj:`numpy.array`
            :param centroids: list of the centroids coordinates.
            :type centroids: :obj:`list`
        """
        distances = self.get_distance(points, centroids)
        return np.argmin(distances, axis=0)

    def get_distance(self, points, centroids):
        """
            Returns the euclidian distance between each point and the centroids.

            :param points: features of each point.
            :type points: :obj:`numpy.array`
            :param centroids: list of the centroids coordinates.
            :type centroids: :obj:`list`
        """
        return np.sqrt(((points - centroids[:, np.newaxis])**2).sum(axis=2))

    def move_centroids(self, points, closest, centroids):
        """
            Returns the new centroids assigned from the points closest to them.

            :param points: features of each point.
            :type points: :obj:`numpy.array`
            :param closest: array with the index of closest centroid for each point.
            :type closest: :obj:`numpy.array`
            :param centroids: list of the centroids coordinates.
            :type centroids: :obj:`list`
        """
        return np.array([points[closest == k].mean(axis=0) for k in range(centroids.shape[0])])

    def silhouette_score(self, X, labels):
        """
            To be added soon
        """
        pass


class GaussianMixture:

    def __init__(self):
        """
            Class for a gaussian mixture model, uses the EM algorithm to fit the model to the data.

            .. warning:: This function is still under development.
                This is a beta version, please be aware that some functionalitie might not be available.
                The full stable version soon will be released.

            :references: * Murphy, K. P. (2012). Machine learning: a probabilistic perspective. MIT press.

            :source: * Inspired by: https://github.com/ocontreras309/ML_Notebooks/blob/master/GMM_Implementation.ipynb
        """
        self.clusters = []
        self.likelihoods = None
        self.likelihood = np.Inf
        self.history = None
        self.scores = None

    def _initialize(self, X, k):
        """
            Initialize the clusters using Kmeans

            :param X: Input data.
            :type X: :obj:`numpy.array`

            :param k: Number of (gaussian) clusters.
            :type k: :obj:`int`
        """
        kmeans = KMeans(k=k)
        kmeans.fit(X)
        means = kmeans.centroids
        for _k in range(k):
            self.clusters.append({
                'pi_k': 1 / k,  # Assign equal probabilities
                'mu_k': means[_k],
                'cov_k': np.identity(X.shape[1], dtype=np.float64)
            })

    def _expect(self, X):
        """
            Expectation step, computes the responsibilities using the means and covariances
            computed at the previous maximization step

            :param X: data
            :type X: :obj:`numpy.array`
        """
        totals = np.zeros((X.shape[0], 1), dtype=np.float64)
        for cluster in self.clusters:
            pi_k = cluster['pi_k']
            mu_k = cluster['mu_k']
            cov_k = cluster['cov_k']

            gamma_nk = (pi_k * gaussian(X, mu_k, cov_k)).astype(np.float64)
            cluster['gamma_nk'] = gamma_nk
            totals += gamma_nk

        for cluster in self.clusters:
            cluster['totals'] = totals
            cluster['gamma_nk'] /= totals

    def _maximize(self, X):
        """
            Maximization step, computes the new values of the means and the covariances

            :param X: data
            :type X: :obj:`numpy.array`
        """
        N = X.shape[0]
        for cluster in self.clusters:
            gamma_nk = cluster['gamma_nk']
            Nk = gamma_nk.sum(axis=0)[0]

            cluster['pi_k'] = Nk / N

            mu_k = (1 / Nk) * (gamma_nk * X).sum(axis=0)
            diff = (X - mu_k).reshape(-1, 1)

            cov_k = np.zeros((X.shape[1], X.shape[1]))
            for j in range(X.shape[0]):
                diff = (X[j] - mu_k).reshape(-1, 1)
                cov_k += gamma_nk[j] * np.dot(diff, diff.T)
            cluster['cov_k'] = cov_k / Nk
            cluster['mu_k'] = mu_k

    def _get_likelihood(self, X):
        """
            Returns the total likelihoods over the dataset

            :param X: data
            :type X: :obj:`numpy.array`
        """
        likelihood = []
        sample_likelihoods = np.log(np.array([cluster['totals'] for cluster in self.clusters]))
        self.likelihood = np.sum(sample_likelihoods)
        return self.likelihood

    def fit(self, X, k, n_epochs=100, improvement_threshold=0.0005):
        """
            Fitting function initialized by K-means algorithm.

            :param X: data.
            :type X: :obj:`numpy.ndarray`
            :param K: number of clusters (gaussians).
            :type K: :obj:`int`
            :param n_epochs: number of epochs, default is 100.
            :type n_epochs: :obj:`numpy.ndarray`
            :param improvement_threshold: Threshold from which we consider the likelihood improved, defaults to 0.0005.
            :type improvement_threshold: :obj:`float`, optional
        """
        self._initialize(X, k)
        likelihoods = np.zeros(n_epochs)
        scores = np.zeros((X.shape[0], k))
        history = []

        diff = 1.
        i = 0

        while (i < n_epochs) and (np.abs(diff) >= improvement_threshold):
            i += 1
            clusters_snapshot = []
            for cluster in self.clusters:
                clusters_snapshot.append({
                    'mu_k': cluster['mu_k'].copy(),
                    'cov_k': cluster['cov_k'].copy()
                })
            history.append(clusters_snapshot)
            self._expect(X)
            self._maximize(X)

            _likelihood_old = self.likelihood
            self.likelihood = self._get_likelihood(X)
            diff            = _likelihood_old - self.likelihood
            likelihoods[i]  = self.likelihood
            print(f'Epoch : {i}, Likelihood : {self.likelihood}')

        for i, cluster in enumerate(self.clusters):
            scores[:, i] = np.log(cluster['gamma_nk']).reshape(-1)

        self.likelihoods = likelihoods
        self.scores = scores
        self.history = history
